import json
from StringIO import StringIO
from urllib import urlencode

from oauthlib import oauth1
from twisted.internet import reactor
from twisted.python.failure import Failure
from twisted.web.client import (
    Agent, FileBodyProducer, PartialDownloadError, readBody)
from twisted.web.http_headers import Headers

from txtwitter.error import TwitterAPIError
from txtwitter.streamservice import TwitterStreamService


TWITTER_API_URL = 'https://api.twitter.com/1.1/'
TWITTER_STREAM_URL = 'https://stream.twitter.com/1.1/'
TWITTER_USERSTREAM_URL = 'https://userstream.twitter.com/1.1/'


def _extract_partial_response(failure):
    failure.trap(PartialDownloadError)
    return failure.value.response


def _read_body(response):
    """
    Read a response body even if there is no content length.
    """
    return readBody(response).addErrback(_extract_partial_response)


def set_bool_param(params, name, value):
    """
    Set a boolean parameter if applicable.

    :param dict params: A dict containing API call parameters.

    :param str name: The name of the parameter to set.

    :param bool value:
        The value of the parameter. If ``None``, the field will not be set. If
        ``True`` or ``False``, the relevant field in ``params`` will be set to
        ``'true'`` or ``'false'``. Any other value will raise a `ValueError`.

    :returns: ``None``
    """
    if value is None:
        return

    if value is True:
        params[name] = 'true'
    elif value is False:
        params[name] = 'false'
    else:
        raise ValueError("Parameter '%s' must be boolean or None, got %r." % (
            name, value))


def set_str_param(params, name, value):
    """
    Set a string parameter if applicable.

    :param dict params: A dict containing API call parameters.

    :param str name: The name of the parameter to set.

    :param bool value:
        The value of the parameter. If ``None``, the field will not be set. If
        an instance of ``basestring``, the relevant field will be set. Any
        other value will raise a `ValueError`.

    :returns: ``None``
    """
    if value is None:
        return

    if isinstance(value, basestring):
        params[name] = value
    else:
        raise ValueError("Parameter '%s' must be a string or None, got %r." % (
            name, value))


def set_float_param(params, name, value, min=None, max=None):
    """
    Set a float parameter if applicable.

    :param dict params: A dict containing API call parameters.

    :param str name: The name of the parameter to set.

    :param float value:
        The value of the parameter. If ``None``, the field will not be set. If
        an instance of a numeric type or a string that can be turned into a
        ``float``, the relevant field will be set. Any other value will raise a
        `ValueError`.

    :param float min:
        If provided, values less than this will raise ``ValueError``.

    :param float max:
        If provided, values greater than this will raise ``ValueError``.

    :returns: ``None``
    """
    if value is None:
        return

    try:
        value = float(str(value))
    except:
        raise ValueError(
            "Parameter '%s' must be numeric (or a numeric string) or None,"
            " got %r." % (name, value))
    if min is not None and value < min:
        raise ValueError(
            "Parameter '%s' must not be less than %r, got %r." % (
                name, min, value))
    if max is not None and value > max:
        raise ValueError(
            "Parameter '%s' must not be greater than %r, got %r." % (
                name, min, value))

    params[name] = str(value)


def set_int_param(params, name, value, min=None, max=None):
    """
    Set a int parameter if applicable.

    :param dict params: A dict containing API call parameters.

    :param str name: The name of the parameter to set.

    :param int value:
        The value of the parameter. If ``None``, the field will not be set. If
        an instance of a numeric type or a string that can be turned into a
        ``int``, the relevant field will be set. Any other value will raise a
        `ValueError`.

    :param int min:
        If provided, values less than this will raise ``ValueError``.

    :param int max:
        If provided, values greater than this will raise ``ValueError``.

    :returns: ``None``
    """
    if value is None:
        return

    try:
        value = int(str(value))
    except:
        raise ValueError(
            "Parameter '%s' must be an integer (or a string representation of"
            " an integer) or None, got %r." % (name, value))
    if min is not None and value < min:
        raise ValueError(
            "Parameter '%s' must not be less than %r, got %r." % (
                name, min, value))
    if max is not None and value > max:
        raise ValueError(
            "Parameter '%s' must not be greater than %r, got %r." % (
                name, min, value))

    params[name] = str(value)


class TwitterClient(object):
    """
    TODO: Document this.
    """
    reactor = reactor

    def __init__(self, token_key, token_secret, consumer_key, consumer_secret,
                 api_url=TWITTER_API_URL, stream_url=TWITTER_STREAM_URL,
                 userstream_url=TWITTER_USERSTREAM_URL, agent=None):
        self._token_key = token_key
        self._token_secret = token_secret
        self._consumer_key = consumer_key
        self._consumer_secret = consumer_secret
        self._api_url_base = api_url
        self._stream_url_base = stream_url
        self._userstream_url_base = userstream_url
        if agent is None:
            agent = Agent(self.reactor)
        self._agent = agent

    def _make_request(self, method, uri, body_parameters=None):
        headers = {}
        body = None
        if body_parameters is not None:
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded',
            }
            body = urlencode(body_parameters)
        client = oauth1.Client(
            self._consumer_key, client_secret=self._consumer_secret,
            resource_owner_key=self._token_key,
            resource_owner_secret=self._token_secret, encoding='utf-8',
            decoding='utf-8')
        uri, headers, body = client.sign(
            uri, http_method=method, headers=headers, body=body)
        headers = Headers(dict((k, [v]) for k, v in headers.items()))

        body_producer = None
        if body is not None:
            body_producer = FileBodyProducer(StringIO(body))

        d = self._agent.request(method, uri, headers, body_producer)
        return d.addCallback(self._handle_error)

    def _handle_error(self, response):
        if response.code < 400:
            return response

        return _read_body(response).addCallback(lambda body: Failure(
            TwitterAPIError(response.code, response=body)))

    def _parse_response(self, response):
        # TODO: Better exception than this.
        assert response.code in (200, 201)
        return readBody(response).addCallback(json.loads)

    def _make_uri(self, base_uri, resource, parameters=None):
        uri = "%s/%s" % (base_uri.rstrip('/'), resource.lstrip('/'))
        if parameters is not None:
            uri = "%s?%s" % (uri, urlencode(parameters))
        return uri

    def _get_api(self, resource, parameters):
        uri = self._make_uri(self._api_url_base, resource, parameters)
        d = self._make_request('GET', uri)
        return d.addCallback(self._parse_response)

    def _post_api(self, resource, parameters):
        uri = self._make_uri(self._api_url_base, resource)
        d = self._make_request('POST', uri, parameters)
        return d.addCallback(self._parse_response)

    def _post_stream(self, resource, parameters):
        uri = self._make_uri(self._stream_url_base, resource)
        return self._make_request('POST', uri, parameters)

    def _get_userstream(self, resource, parameters):
        uri = self._make_uri(self._userstream_url_base, resource, parameters)
        return self._make_request('GET', uri)

    # Timelines

    def statuses_mentions_timeline(self, count=None, since_id=None,
                                   max_id=None, trim_user=None,
                                   contributor_details=None,
                                   include_entities=None):
        """
        Returns a list of the most recent mentions (tweets containing a users's
        @screen_name) for the authenticating user.

        https://dev.twitter.com/docs/api/1.1/get/statuses/mentions_timeline

        :param int count:
            Specifies the number of tweets to try and retrieve, up to a maximum
            of 200.

        :param str since_id:
            Returns results with an ID greater than (that is, more recent than)
            the specified ID. Tweets newer than this may not be returned due to
            certain API limits.

        :param str max_id:
            Returns results with an ID less than (that is, older than) or equal
            to the specified ID.

        :param bool trim_user:
            When set to ``True``, the tweet's user object includes only the
            status author's numerical ID.

        :param bool contributor_details:
            This parameter enhances the contributors element of the status
            response to include the screen_name of the contributor. By default
            only the user_id of the contributor is included.

        :param bool include_entities:
            When set to ``False``, the ``entities`` node will not be included.

        :returns: A list of tweet dicts.
        """
        params = {}
        set_int_param(params, 'count', count)
        set_str_param(params, 'since_id', since_id)
        set_str_param(params, 'max_id', max_id)
        set_bool_param(params, 'trim_user', trim_user)
        set_bool_param(params, 'contributor_details', contributor_details)
        set_bool_param(params, 'include_entities', include_entities)
        return self._get_api('statuses/mentions_timeline.json', params)

    def statuses_user_timeline(self, user_id=None, screen_name=None,
                               since_id=None, count=None, max_id=None,
                               trim_user=None, exclude_replies=None,
                               contributor_details=None,
                               include_rts=None):
        """
        Returns a list of the most recent tweets posted by the specified user.

        https://dev.twitter.com/docs/api/1.1/get/statuses/user_timeline

        Either ``user_id`` or ``screen_name`` must be provided.

        :param str user_id:
            The ID of the user to return tweets for.

        :param str screen_name:
            The screen name of the user to return tweets for.

        :param str since_id:
            Returns results with an ID greater than (that is, more recent than)
            the specified ID. Tweets newer than this may not be returned due to
            certain API limits.

        :param int count:
            Specifies the number of tweets to try and retrieve, up to a maximum
            of 200.

        :param str max_id:
            Returns results with an ID less than (that is, older than) or equal
            to the specified ID.

        :param bool trim_user:
            When set to ``True``, the tweet's user object includes only the
            status author's numerical ID.

        :param bool exclude_replies:
            When set to ``True``, replies will not appear in the timeline.

        :param bool contributor_details:
            This parameter enhances the contributors element of the status
            response to include the screen_name of the contributor. By default
            only the user_id of the contributor is included.

        :param bool include_rts:
            When set to ``False``, retweets will not appear in the timeline.

        :returns: A list of tweet dicts.
        """
        params = {}
        set_str_param(params, 'user_id', user_id)
        set_str_param(params, 'screen_name', screen_name)
        set_str_param(params, 'since_id', since_id)
        set_int_param(params, 'count', count)
        set_str_param(params, 'max_id', max_id)
        set_bool_param(params, 'trim_user', trim_user)
        set_bool_param(params, 'exclude_replies', exclude_replies)
        set_bool_param(params, 'contributor_details', contributor_details)
        set_bool_param(params, 'include_rts', include_rts)
        return self._get_api('statuses/user_timeline.json', params)

    def statuses_home_timeline(self, count=None, since_id=None, max_id=None,
                               trim_user=None, exclude_replies=None,
                               contributor_details=None,
                               include_entities=None):
        """
        Returns a collection of the most recent Tweets and retweets posted by
        the authenticating user and the users they follow.

        https://dev.twitter.com/docs/api/1.1/get/statuses/home_timeline

        :param int count:
            Specifies the number of tweets to try and retrieve, up to a maximum
            of 200.

        :param str since_id:
            Returns results with an ID greater than (that is, more recent than)
            the specified ID. Tweets newer than this may not be returned due to
            certain API limits.

        :param str max_id:
            Returns results with an ID less than (that is, older than) or equal
            to the specified ID.

        :param bool trim_user:
            When set to ``True``, the tweet's user object includes only the
            status author's numerical ID.

        :param bool exclude_replies:
            When set to ``True``, replies will not appear in the timeline.

        :param bool contributor_details:
            This parameter enhances the contributors element of the status
            response to include the screen_name of the contributor. By default
            only the user_id of the contributor is included.

        :param bool include_entities:
            When set to ``False``, the ``entities`` node will not be included.

        :returns: A list of tweet dicts.
        """
        params = {}
        set_int_param(params, 'count', count)
        set_str_param(params, 'since_id', since_id)
        set_str_param(params, 'max_id', max_id)
        set_bool_param(params, 'trim_user', trim_user)
        set_bool_param(params, 'exclude_replies', exclude_replies)
        set_bool_param(params, 'contributor_details', contributor_details)
        set_bool_param(params, 'include_entities', include_entities)
        return self._get_api('statuses/home_timeline.json', params)

    # TODO: Implement statuses_retweets_of_me()

    # Tweets

    def statuses_retweets(self, id, count=None, trim_user=None):
        """
        Returns a list of the most recent retweets of the Tweet specified by
        the id parameter.

        https://dev.twitter.com/docs/api/1.1/get/statuses/retweets/%3Aid

        :param str id:
            (*required*) The numerical ID of the desired tweet.

        :param int count:
            The maximum number of retweets to return. (Max 100)

        :param bool trim_user:
            When set to ``True``, the tweet's user object includes only the
            status author's numerical ID.

        :returns: A tweet dict.
        """
        params = {'id': id}
        set_int_param(params, 'count', count)
        set_bool_param(params, 'trim_user', trim_user)
        return self._get_api('statuses/retweets.json', params)

    def statuses_show(self, id, trim_user=None, include_my_retweet=None,
                      include_entities=None):
        """
        Returns a single Tweet, specified by the id parameter.

        https://dev.twitter.com/docs/api/1.1/get/statuses/show/%3Aid

        :param str id:
            (*required*) The numerical ID of the desired tweet.

        :param bool trim_user:
            When set to ``True``, the tweet's user object includes only the
            status author's numerical ID.

        :param bool include_my_retweet:
            When set to ``True``, any Tweet returned that has been retweeted by
            the authenticating user will include an additional
            ``current_user_retweet`` node, containing the ID of the source
            status for the retweet.

        :param bool include_entities:
            When set to ``False``, the ``entities`` node will not be included.

        :returns: A tweet dict.
        """
        params = {'id': id}
        set_bool_param(params, 'trim_user', trim_user)
        set_bool_param(params, 'include_my_retweet', include_my_retweet)
        set_bool_param(params, 'include_entities', include_entities)
        return self._get_api('statuses/show.json', params)

    def statuses_destroy(self, id, trim_user=None):
        """
        Destroys the status specified by the ID parameter.

        https://dev.twitter.com/docs/api/1.1/post/statuses/destroy/%3Aid

        :param str id:
            (*required*) The numerical ID of the desired tweet.

        :param bool trim_user:
            When set to ``True``, the return value's user object includes only
            the status author's numerical ID.

        :returns:
            A tweet dict containing the destroyed tweet.
        """
        params = {'id': id}
        set_bool_param(params, 'trim_user', trim_user)
        return self._post_api('statuses/destroy.json', params)

    def statuses_update(self, status, in_reply_to_status_id=None, lat=None,
                        long=None, place_id=None, display_coordinates=None,
                        trim_user=None):
        """
        Posts a tweet.

        https://dev.twitter.com/docs/api/1.1/post/statuses/update

        :param str status:
            (*required*) The text of your tweet, typically up to 140
            characters. URL encode as necessary. t.co link wrapping may affect
            character counts.

            There are some special commands in this field to be aware of. For
            instance, preceding a message with "D " or "M " and following it
            with a screen name can create a direct message to that user if the
            relationship allows for it.

        :param str in_reply_to_status_id:
            The ID of an existing status that the update is in reply to.

            Note: This parameter will be ignored unless the author of the tweet
            this parameter references is mentioned within the status text.
            Therefore, you must include @username, where username is the author
            of the referenced tweet, within ``status``.

        :param float lat:
            The latitude of the location this tweet refers to. This parameter
            will be ignored unless it is inside the range -90.0 to +90.0 (North
            is positive) inclusive. It will also be ignored if there isn't a
            corresponding long parameter.

        :param float long:
            The longitude of the location this tweet refers to. The valid
            ranges for longitude is -180.0 to +180.0 (East is positive)
            inclusive. This parameter will be ignored if outside that range, if
            it is not a number, if geo_enabled is disabled, or if there not a
            corresponding lat parameter.

        :param str place_id:
            A place in the world. These IDs can be retrieved from GET
            geo/reverse_geocode. (TODO: Reference method when it exists.)

        :param bool display_coordinates:
            Whether or not to put a pin on the exact coordinates a tweet has
            been sent from.

        :param bool trim_user:
            When set to ``True``, the return value's user object includes only
            the status author's numerical ID.

        :returns:
            A tweet dict containing the posted tweet.
        """
        params = {'status': status}
        set_str_param(params, 'in_reply_to_status_id', in_reply_to_status_id)
        set_float_param(params, 'lat', lat, min=-90, max=90)
        set_float_param(params, 'long', long, min=-180, max=180)
        set_str_param(params, 'place_id', place_id)
        set_bool_param(params, 'display_coordinates', display_coordinates)
        set_bool_param(params, 'trim_user', trim_user)
        return self._post_api('statuses/update.json', params)

    def statuses_retweet(self, id, trim_user=None):
        """
        Retweets the status specified by the ID parameter.

        https://dev.twitter.com/docs/api/1.1/post/statuses/retweet/%3Aid

        :param str id:
            (*required*) The numerical ID of the desired tweet.

        :param bool trim_user:
            When set to ``True``, the return value's user object includes only
            the status author's numerical ID.

        :returns:
            A tweet dict containing the retweet. (Contains the retweeted tweet
            in the ``retweeted_status`` field.)
        """
        params = {'id': id}
        set_bool_param(params, 'trim_user', trim_user)
        return self._post_api('statuses/retweet.json', params)

    # TODO: Implement statuses_update_with_media()
    # TODO: Implement statuses_oembed()
    # TODO: Implement statuses_retweeters_ids()

    # Search

    # TODO: Implement search_tweets()

    # Streaming

    def stream_filter(self, delegate, follow=None, track=None, locations=None,
                      stall_warnings=None):
        """
        Streams public messages filtered by various parameters.

        https://dev.twitter.com/docs/api/1.1/post/statuses/filter

        At least one of ``follow``, ``track``, or ``locations`` must be
        provided. See the API documentation linked above for details on these
        parameters and the various limits on this API.

        :param delegate:
            A delegate function that will be called for each message in the
            stream and will be passed the message dict as the only parameter.
            The message dicts passed to this function may represent any message
            type and the delegate is responsible for any dispatch that may be
            required. (:mod:`txtwitter.messagetools` may be helpful here.)

        :param list follow:
            A list of user IDs, indicating the users to return statuses for in
            the stream.

        :param list track:
            List of keywords to track.

        :param list locations:
            List of location bounding boxes to track.
            XXX: Currently unsupported.

        :param bool stall_warnings:
            Specifies whether stall warnings should be delivered.

        :returns: An unstarted :class:`TwitterStreamService`.
        """
        params = {}
        if follow is not None:
            params['follow'] = ','.join(follow)
        if track is not None:
            params['track'] = ','.join(track)
        if locations is not None:
            raise NotImplementedError(
                "The `locations` parameter is not yet supported.")
        set_bool_param(params, 'stall_warnings', stall_warnings)

        svc = TwitterStreamService(
            lambda: self._post_stream('statuses/filter.json', params),
            delegate)
        return svc

    # TODO: Implement stream_sample()
    # TODO: Implement stream_firehose()

    def userstream_user(self, delegate, stall_warnings=None,
                        with_='followings', replies=None):
        """
        Streams messages for a single user.

        https://dev.twitter.com/docs/api/1.1/get/user

        The ``stringify_friend_ids`` parameter is always set to ``'true'`` for
        consistency with the use of string identifiers elsewhere.

        :param delegate:
            A delegate function that will be called for each message in the
            stream and will be passed the message dict as the only parameter.
            The message dicts passed to this function may represent any message
            type and the delegate is responsible for any dispatch that may be
            required. (:mod:`txtwitter.messagetools` may be helpful here.)

        :param bool stall_warnings:
            Specifies whether stall warnings should be delivered.

        :param str with_:
            If ``'followings'`` (the default), the stream will include messages
            from both the authenticated user and the authenticated user's
            followers. If ``'user'``, the stream will only include messages
            from (or mentioning) the autheticated user. All other values are
            invalid. (The underscore appended to the parameter name is to avoid
            conflicting with Python's ``with`` keyword.)

        :param str replies:
            If set to ``'all'``, replies to tweets will be included even if the
            authenticated user does not follow both parties.

        :returns: An unstarted :class:`TwitterStreamService`.
        """
        params = {'stringify_friend_ids': 'true'}
        set_bool_param(params, 'stall_warnings', stall_warnings)
        set_str_param(params, 'with', with_)
        set_str_param(params, 'replies', replies)

        svc = TwitterStreamService(
            lambda: self._get_userstream('user.json', params),
            delegate)
        return svc

    # Direct Messages

    def direct_messages(self, since_id=None, max_id=None, count=None,
                        include_entities=None, skip_status=None):
        """
        Gets the 20 most recent direct messages received by the authenticating
        user.

        https://dev.twitter.com/docs/api/1.1/get/direct_messages

        :param str since_id:
            Returns results with an ID greater than (that is, more recent than)
            the specified ID. There are limits to the number of Tweets which
            can be accessed through the API. If the limit of Tweets has occured
            since the since_id, the since_id will be forced to the oldest ID
            available.

        :params str max_id:
            Returns results with an ID less than (that is, older than) or equal
            to the specified ID.

        :param int count:
            Specifies the number of direct messages to try and retrieve, up to
            a maximum of ``200``. The value of count is best thought of as a
            limit to the number of Tweets to return because suspended or
            deleted content is removed after the count has been applied.

        :param bool include_entities:
            The entities node will not be included when set to ``False``.

        :param bool skip_status:
            When set to ``True``, statuses will not be included in the returned
            user objects.

        :returns:
            A list of direct message dicts.
        """
        params = {}
        set_str_param(params, 'since_id', since_id)
        set_str_param(params, 'max_id', max_id)
        set_int_param(params, 'count', count)
        set_bool_param(params, 'include_entities', include_entities)
        set_bool_param(params, 'skip_status', skip_status)
        return self._get_api('direct_messages.json', params)

    def direct_messages_sent(self, since_id=None, max_id=None, count=None,
                             include_entities=None, page=None):
        """
        Gets the 20 most recent direct messages sent by the authenticating
        user.

        https://dev.twitter.com/docs/api/1.1/get/direct_messages/sent

        :param str since_id:
            Returns results with an ID greater than (that is, more recent than)
            the specified ID. There are limits to the number of Tweets which
            can be accessed through the API. If the limit of Tweets has occured
            since the since_id, the since_id will be forced to the oldest ID
            available.

        :params str max_id:
            Returns results with an ID less than (that is, older than) or equal
            to the specified ID.

        :param int count:
            Returns results with an ID less than (that is, older than) or equal
            to the specified ID.

        :param int page:
            Specifies the page of results to retrieve.

        :param bool include_entities:
            The entities node will not be included when set to ``False``.

        :returns:
            A list of direct message dicts.
        """
        params = {}
        set_str_param(params, 'since_id', since_id)
        set_str_param(params, 'max_id', max_id)
        set_int_param(params, 'count', count)
        set_int_param(params, 'page', page)
        set_bool_param(params, 'include_entities', include_entities)
        return self._get_api('direct_messages/sent.json', params)

    def direct_messages_show(self, id):
        """
        Gets the direct message with the given id.

        https://dev.twitter.com/docs/api/1.1/get/direct_messages/show

        :param str id:
            (*required*) The ID of the direct message.

        :returns:
            A direct message dict.
        """
        params = {}
        set_str_param(params, 'id', id)
        d = self._get_api('direct_messages/show.json', params)
        d.addCallback(lambda dms: dms[0])
        return d

    def direct_messages_destroy(self, id, include_entities=None):
        """
        Destroys the direct message with the given id.

        https://dev.twitter.com/docs/api/1.1/post/direct_messages/destroy

        :param str id:
            (*required*) The ID of the direct message.
        :param bool include_entities:
            The entities node will not be included when set to ``False``.

        :returns:
            A direct message dict containing the destroyed direct message.
        """
        params = {}
        set_str_param(params, 'id', id)
        set_bool_param(params, 'include_entities', include_entities)
        return self._post_api('direct_messages/destroy.json', params)

    def direct_messages_new(self, text, user_id=None, screen_name=None):
        """
        Sends a new direct message to the given user from the authenticating
        user.

        https://dev.twitter.com/docs/api/1.1/post/direct_messages/new

        :param str text:
            (*required*) The text of your direct message.
        :param str user_id:
            The ID of the user who should receive the direct message. Required
            if ``screen_name`` isn't given.
        :param str screen_name:
            The screen name of the user who should receive the direct message.
            Required if ``user_id`` isn't given.

        :returns:
            A direct message dict containing the sent direct message.
        """
        params = {}
        set_str_param(params, 'text', text)
        set_str_param(params, 'user_id', user_id)
        set_str_param(params, 'screen_name', screen_name)
        return self._post_api('direct_messages/new.json', params)

    # TODO: Implement direct_messages_show()
    # TODO: Implement direct_messages_destroy()
    # TODO: Implement direct_messages_new()

    # Friends & Followers

    # TODO: Implement friendships_no_retweets_ids()
    # TODO: Implement friends_ids()
    # TODO: Implement followers_ids()
    # TODO: Implement friendships_lookup()
    # TODO: Implement friendships_incoming()
    # TODO: Implement friendships_outgoing()
    # TODO: Implement friendships_create()
    # TODO: Implement friendships_destroy()
    # TODO: Implement friendships_update()
    # TODO: Implement friendships_show()
    # TODO: Implement friends_list()
    # TODO: Implement followers_list()

    # Users

    # TODO: Implement account_settings()
    # TODO: Implement account_verify_credentials()
    # TODO: Implement account_settings()
    # TODO: Implement account_update_delivery_device()
    # TODO: Implement account_update_profile()
    # TODO: Implement account_update_profile_background_image()
    # TODO: Implement account_update_profile_colors()
    # TODO: Implement account_update_profile_image()
    # TODO: Implement blocks_list()
    # TODO: Implement blocks_ids()
    # TODO: Implement blocks_create()
    # TODO: Implement blocks_destroy()
    # TODO: Implement users_lookup()
    # TODO: Implement users_show()
    # TODO: Implement users_search()
    # TODO: Implement users_contributees()
    # TODO: Implement users_contributors()
    # TODO: Implement account_remove_profile_banner()
    # TODO: Implement account_update_profile_banner()
    # TODO: Implement users/profile_banner()

    # Suggested Users

    # TODO: Implement users_suggestions()
    # TODO: Implement users_suggestions()
    # TODO: Implement users_suggestions_members()

    # Favorites

    # TODO: Implement favorites_list()
    # TODO: Implement favorites_destroy()
    # TODO: Implement favorites_create()

    # Lists

    # TODO: Implement lists_list()
    # TODO: Implement lists_statuses()
    # TODO: Implement lists_members_destroy()
    # TODO: Implement lists_memberships()
    # TODO: Implement lists_subscribers()
    # TODO: Implement lists_subscribers/create()
    # TODO: Implement lists_subscribers/show()
    # TODO: Implement lists_subscribers/destroy()
    # TODO: Implement lists_members_create_all()
    # TODO: Implement lists_members_show()
    # TODO: Implement lists_members()
    # TODO: Implement lists_members_create()
    # TODO: Implement lists_destroy()
    # TODO: Implement lists_update()
    # TODO: Implement lists_create()
    # TODO: Implement lists_show()
    # TODO: Implement lists_subscriptions()
    # TODO: Implement lists_members_destroy_all()
    # TODO: Implement lists_ownerships()

    # Saved Searches

    # TODO: Implement saved_searches_list()
    # TODO: Implement saved_searches_show()
    # TODO: Implement saved_searches_create()
    # TODO: Implement saved_searches_destroy()

    # Places & Geo

    # TODO: Implement geo_id()
    # TODO: Implement geo_reverse_geocode()
    # TODO: Implement geo_search()
    # TODO: Implement geo_similar_places()
    # TODO: Implement geo_place()

    # Trends

    # TODO: Implement trends_place()
    # TODO: Implement trends_available()
    # TODO: Implement trends_closest()

    # Spam Reporting

    # TODO: Implement users_report_spam()

    # OAuth

    # TODO: Decide whether any of these APIs should be implemented.

    # Help

    # TODO: Implement help_configuration()
    # TODO: Implement help_languages()
    # TODO: Implement help_privacy()
    # TODO: Implement help_tos()
    # TODO: Implement application_rate_limit_status()



import json

from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.trial.unittest import TestCase

from txtwitter.tests.fake_agent import FakeAgent, FakeResponse


def from_twitter(name):
    @property
    def prop(self):
        from txtwitter import twitter
        return getattr(twitter, name)
    return prop


class TestParamHelpers(TestCase):
    _set_bool_param = from_twitter('set_bool_param')
    _set_str_param = from_twitter('set_str_param')
    _set_float_param = from_twitter('set_float_param')
    _set_int_param = from_twitter('set_int_param')

    def test_set_bool_param_None(self):
        """
        set_bool_param() should do nothing if the value is ``None``.
        """
        params = {}
        self._set_bool_param(params, 'foo', None)
        self.assertEqual(params, {})

    def test_set_bool_param_True(self):
        """
        set_bool_param() should set the param to 'true' if the value is
        ``True``.
        """
        params = {}
        self._set_bool_param(params, 'foo', True)
        self.assertEqual(params, {'foo': 'true'})

    def test_set_bool_param_False(self):
        """
        set_bool_param() should set the param to 'false' if the value is
        ``False``.
        """
        params = {}
        self._set_bool_param(params, 'foo', False)
        self.assertEqual(params, {'foo': 'false'})

    def test_set_bool_param_bad(self):
        """
        set_bool_param() should raise ValueError if the value is not a bool or
        ``None``.
        """
        self.assertRaises(ValueError, self._set_bool_param, {}, 'foo', 1)

    def test_set_str_param_None(self):
        """
        set_str_param() should do nothing if the value is ``None``.
        """
        params = {}
        self._set_str_param(params, 'foo', None)
        self.assertEqual(params, {})

    def test_set_str_param(self):
        """
        set_str_param() should set the param if the value is a string.
        """
        params = {}
        self._set_str_param(params, 'foo', 'some string')
        self.assertEqual(params, {'foo': 'some string'})

    def test_set_str_param_bad(self):
        """
        set_str_param() should raise ValueError if the value is not a string or
        ``None``.
        """
        self.assertRaises(ValueError, self._set_str_param, {}, 'foo', True)

    def test_set_float_param_None(self):
        """
        set_float_param() should do nothing if the value is ``None``.
        """
        params = {}
        self._set_float_param(params, 'foo', None)
        self.assertEqual(params, {})

    def test_set_float_param(self):
        """
        set_float_param() should set the param if the value is a float.
        """
        params = {}
        self._set_float_param(params, 'foo', 42.5)
        self.assertEqual(params, {'foo': '42.5'})

    def test_set_float_param_int(self):
        """
        set_float_param() should set the param if the value is an int.
        """
        params = {}
        self._set_float_param(params, 'foo', 42)
        self.assertEqual(params, {'foo': '42.0'})

    def test_set_float_param_str(self):
        """
        set_float_param() should set the param if the value is a string
        representation of a float.
        """
        params = {}
        self._set_float_param(params, 'foo', '42.5')
        self.assertEqual(params, {'foo': '42.5'})

    def test_set_float_param_min(self):
        """
        set_float_param() should raise ValueError if the value is less than the
        minimum.
        """
        self.assertRaises(
            ValueError, self._set_float_param, {}, 'foo', 42.5, min=50)

    def test_set_float_param_max(self):
        """
        set_float_param() should raise ValueError if the value is greater than
        the maximum.
        """
        self.assertRaises(
            ValueError, self._set_float_param, {}, 'foo', 42.5, max=40)

    def test_set_float_param_min_str(self):
        """
        set_float_param() should raise ValueError if the value is less than the
        minimum.
        """
        self.assertRaises(
            ValueError, self._set_float_param, {}, 'foo', '42.5', min=50)

    def test_set_float_param_max_str(self):
        """
        set_float_param() should raise ValueError if the value is greater than
        the maximum.
        """
        self.assertRaises(
            ValueError, self._set_float_param, {}, 'foo', '42.5', max=40)

    def test_set_float_param_bad(self):
        """
        set_float_param() should raise ValueError if the value is not a float
        (and cannot be turned into one) or ``None``.
        """
        self.assertRaises(ValueError, self._set_float_param, {}, 'foo', True)

    def test_set_int_param_None(self):
        """
        set_int_param() should do nothing if the value is ``None``.
        """
        params = {}
        self._set_int_param(params, 'foo', None)
        self.assertEqual(params, {})

    def test_set_int_param(self):
        """
        set_int_param() should set the param if the value is a int.
        """
        params = {}
        self._set_int_param(params, 'foo', 42)
        self.assertEqual(params, {'foo': '42'})

    def test_set_int_param_str(self):
        """
        set_int_param() should set the param if the value is a string
        representation of an int.
        """
        params = {}
        self._set_int_param(params, 'foo', '42')
        self.assertEqual(params, {'foo': '42'})

    def test_set_int_param_min(self):
        """
        set_int_param() should raise ValueError if the value is less than the
        minimum.
        """
        self.assertRaises(
            ValueError, self._set_int_param, {}, 'foo', 42, min=50)

    def test_set_int_param_max(self):
        """
        set_int_param() should raise ValueError if the value is greater than
        the maximum.
        """
        self.assertRaises(
            ValueError, self._set_int_param, {}, 'foo', 42, max=40)

    def test_set_int_param_min_str(self):
        """
        set_int_param() should raise ValueError if the value is less than the
        minimum.
        """
        self.assertRaises(
            ValueError, self._set_int_param, {}, 'foo', '42', min=50)

    def test_set_int_param_max_str(self):
        """
        set_int_param() should raise ValueError if the value is greater than
        the maximum.
        """
        self.assertRaises(
            ValueError, self._set_int_param, {}, 'foo', '42', max=40)

    def test_set_int_param_bad(self):
        """
        set_int_param() should raise ValueError if the value is not a int
        (and cannot be turned into one) or ``None``.
        """
        self.assertRaises(ValueError, self._set_int_param, {}, 42.5, True)


class TestTwitterClient(TestCase):
    timeout = 1

    _TwitterClient = from_twitter('TwitterClient')

    def _resp_json(self, data, code=200):
        return FakeResponse(json.dumps(data), code)

    def _agent_and_TwitterClient(self):
        agent = FakeAgent()
        client = self._TwitterClient(
            'token-key', 'token-secret', 'consumer-key', 'consumer-secret',
            agent=agent)
        return agent, client

    # Timelines

    @inlineCallbacks
    def test_statuses_mentions_timeline(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/mentions_timeline.json'
        response_list = [{
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }, {
            # Truncated tweet data.
            "id_str": "122",
            "text": "Tweet!",
        }]
        agent.add_expected_request(
            'GET', uri, {}, self._resp_json(response_list))
        resp = yield client.statuses_mentions_timeline()
        self.assertEqual(resp, response_list)

    @inlineCallbacks
    def test_statuses_mentions_timeline_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/mentions_timeline.json'
        response_list = [{
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }, {
            # Truncated tweet data.
            "id_str": "122",
            "text": "Tweet!",
        }]
        expected_params = {
            'count': '10',
            'since_id': '121',
            'max_id': '123',
            'trim_user': 'true',
            'contributor_details': 'true',
            'include_entities': 'false',
        }
        agent.add_expected_request(
            'GET', uri, expected_params, self._resp_json(response_list))
        resp = yield client.statuses_mentions_timeline(
            count=10, since_id='121', max_id='123', trim_user=True,
            contributor_details=True, include_entities=False)
        self.assertEqual(resp, response_list)

    @inlineCallbacks
    def test_statuses_user_timeline(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/user_timeline.json'
        response_list = [{
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }, {
            # Truncated tweet data.
            "id_str": "122",
            "text": "Tweet!",
        }]
        agent.add_expected_request(
            'GET', uri, {}, self._resp_json(response_list))
        resp = yield client.statuses_user_timeline()
        self.assertEqual(resp, response_list)

    @inlineCallbacks
    def test_statuses_user_timeline_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/user_timeline.json'
        response_list = [{
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }, {
            # Truncated tweet data.
            "id_str": "122",
            "text": "Tweet!",
        }]
        expected_params = {
            'user_id': '12',
            'screen_name': 'fakeuser',
            'since_id': '121',
            'count': '10',
            'max_id': '123',
            'trim_user': 'true',
            'exclude_replies': 'true',
            'contributor_details': 'true',
            'include_rts': 'false',
        }
        agent.add_expected_request(
            'GET', uri, expected_params, self._resp_json(response_list))
        resp = yield client.statuses_user_timeline(
            user_id='12', screen_name='fakeuser', since_id='121', count=10,
            max_id='123', trim_user=True, exclude_replies=True,
            contributor_details=True, include_rts=False)
        self.assertEqual(resp, response_list)

    @inlineCallbacks
    def test_statuses_home_timeline(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/home_timeline.json'
        response_list = [{
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }, {
            # Truncated tweet data.
            "id_str": "122",
            "text": "Tweet!",
        }]
        agent.add_expected_request(
            'GET', uri, {}, self._resp_json(response_list))
        resp = yield client.statuses_home_timeline()
        self.assertEqual(resp, response_list)

    @inlineCallbacks
    def test_statuses_home_timeline_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/home_timeline.json'
        response_list = [{
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }, {
            # Truncated tweet data.
            "id_str": "122",
            "text": "Tweet!",
        }]
        expected_params = {
            'count': '10',
            'since_id': '121',
            'max_id': '123',
            'trim_user': 'true',
            'exclude_replies': 'true',
            'contributor_details': 'true',
            'include_entities': 'false',
        }
        agent.add_expected_request(
            'GET', uri, expected_params, self._resp_json(response_list))
        resp = yield client.statuses_home_timeline(
            count=10, since_id='121', max_id='123', trim_user=True,
            exclude_replies=True, contributor_details=True,
            include_entities=False)
        self.assertEqual(resp, response_list)

    # TODO: Tests for statuses_retweets_of_me()

    # Tweets

    @inlineCallbacks
    def test_statuses_retweets(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/retweets.json'
        response_list = [{
            # Truncated tweet data.
            "id_str": "124",
            "text": "Tweet!",
            "retweeted_status": {
                "id_str": "123",
                "text": "Tweet!",
            },
        }, {
            # Truncated tweet data.
            "id_str": "125",
            "text": "Tweet!",
            "retweeted_status": {
                "id_str": "123",
                "text": "Tweet!",
            },
        }]
        agent.add_expected_request(
            'GET', uri, {'id': '123'}, self._resp_json(response_list))
        resp = yield client.statuses_retweets("123")
        self.assertEqual(resp, response_list)

    @inlineCallbacks
    def test_statuses_retweets_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/retweets.json'
        response_list = [{
            # Truncated tweet data.
            "id_str": "124",
            "text": "Tweet!",
            "retweeted_status": {
                "id_str": "123",
                "text": "Tweet!",
            },
        }, {
            # Truncated tweet data.
            "id_str": "125",
            "text": "Tweet!",
            "retweeted_status": {
                "id_str": "123",
                "text": "Tweet!",
            },
        }]
        expected_params = {
            'id': '123',
            'count': '10',
            'trim_user': 'true',
        }
        agent.add_expected_request(
            'GET', uri, expected_params, self._resp_json(response_list))
        resp = yield client.statuses_retweets("123", count=10, trim_user=True)
        self.assertEqual(resp, response_list)

    @inlineCallbacks
    def test_statuses_show(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/show.json'
        response_dict = {
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }
        agent.add_expected_request(
            'GET', uri, {'id': '123'}, self._resp_json(response_dict))
        resp = yield client.statuses_show("123")
        self.assertEqual(resp, response_dict)

    @inlineCallbacks
    def test_statuses_show_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/show.json'
        response_dict = {
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }
        expected_params = {
            'id': '123',
            'trim_user': 'true',
            'include_my_retweet': 'true',
            'include_entities': 'false',
        }
        agent.add_expected_request(
            'GET', uri, expected_params, self._resp_json(response_dict))
        resp = yield client.statuses_show(
            "123", trim_user=True, include_my_retweet=True,
            include_entities=False)
        self.assertEqual(resp, response_dict)

    @inlineCallbacks
    def test_statuses_show_HTTP_404(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/show.json'
        err_dict = {"errors": [
            {"message": "Sorry, that page does not exist", "code": 34},
        ]}
        agent.add_expected_request(
            'GET', uri, {'id': '123'}, self._resp_json(err_dict, 404))
        err = yield client.statuses_show("123").addErrback(lambda f: f.value)
        code, _phrase, body = err.args
        self.assertEqual((404, err_dict), (code, json.loads(body)))

    @inlineCallbacks
    def test_statuses_destroy(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/destroy.json'
        response_dict = {
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }
        agent.add_expected_request(
            'POST', uri, {'id': '123'}, self._resp_json(response_dict))
        resp = yield client.statuses_destroy("123")
        self.assertEqual(resp, response_dict)

    @inlineCallbacks
    def test_statuses_destroy_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/destroy.json'
        response_dict = {
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }
        expected_params = {
            'id': '123',
            'trim_user': 'true',
        }
        agent.add_expected_request(
            'POST', uri, expected_params, self._resp_json(response_dict))
        resp = yield client.statuses_destroy("123", trim_user=True)
        self.assertEqual(resp, response_dict)

    @inlineCallbacks
    def test_statuses_update(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/update.json'
        response_dict = {
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }
        agent.add_expected_request(
            'POST', uri, {'status': 'Tweet!'}, self._resp_json(response_dict))
        resp = yield client.statuses_update("Tweet!")
        self.assertEqual(resp, response_dict)

    @inlineCallbacks
    def test_statuses_update_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/update.json'
        response_dict = {
            # Truncated tweet data.
            "id_str": "123",
            "text": "Tweet!",
        }
        expected_params = {
            'status': 'Tweet!',
            'in_reply_to_status_id': '122',
            'lat': '-33.93',
            'long': '18.42',
            'place_id': 'abc123',
            'display_coordinates': 'true',
            'trim_user': 'true',
        }
        agent.add_expected_request(
            'POST', uri, expected_params, self._resp_json(response_dict))
        resp = yield client.statuses_update(
            "Tweet!", in_reply_to_status_id="122", lat=-33.93, long=18.42,
            place_id='abc123', display_coordinates=True, trim_user=True)
        self.assertEqual(resp, response_dict)

    @inlineCallbacks
    def test_statuses_retweet(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/retweet.json'
        response_dict = {
            # Truncated tweet data.
            "id_str": "124",
            "text": "Tweet!",
            "retweeted_status": {
                "id_str": "123",
                "text": "Tweet!",
            },
        }
        agent.add_expected_request(
            'POST', uri, {'id': '123'}, self._resp_json(response_dict))
        resp = yield client.statuses_retweet("123")
        self.assertEqual(resp, response_dict)

    @inlineCallbacks
    def test_statuses_retweet_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/statuses/retweet.json'
        response_dict = {
            # Truncated tweet data.
            "id_str": "124",
            "text": "Tweet!",
            "retweeted_status": {
                "id_str": "123",
                "text": "Tweet!",
            },
        }
        expected_params = {
            'id': '123',
            'trim_user': 'true',
        }
        agent.add_expected_request(
            'POST', uri, expected_params, self._resp_json(response_dict))
        resp = yield client.statuses_retweet("123", trim_user=True)
        self.assertEqual(resp, response_dict)

    # TODO: Tests for statuses_update_with_media()
    # TODO: Tests for statuses_oembed()
    # TODO: Tests for statuses_retweeters_ids()

    # Search

    # TODO: Tests for search_tweets()

    # Streaming

    @inlineCallbacks
    def test_stream_filter_track(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://stream.twitter.com/1.1/statuses/filter.json'
        stream = FakeResponse(None)
        agent.add_expected_request('POST', uri, {'track': 'foo,bar'}, stream)

        connected = Deferred()
        tweets = []
        svc = client.stream_filter(tweets.append, track=['foo', 'bar'])
        svc.set_connect_callback(connected.callback)
        svc.startService()
        connected_svc = yield connected
        self.assertIs(svc, connected_svc)
        self.assertEqual(tweets, [])

        stream.deliver_data(
            '{"id_str": "1", "text": "Tweet 1", "user": {}}\r\n')
        self.assertEqual(tweets, [
            {"id_str": "1", "text": "Tweet 1", "user": {}},
        ])
        stream.deliver_data(
            '{"id_str": "2", "text": "Tweet 2", "user": {}}\r\n')
        self.assertEqual(tweets, [
            {"id_str": "1", "text": "Tweet 1", "user": {}},
            {"id_str": "2", "text": "Tweet 2", "user": {}},
        ])
        yield svc.stopService()
        stream.finished()

    @inlineCallbacks
    def test_stream_filter_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://stream.twitter.com/1.1/statuses/filter.json'
        stream = FakeResponse(None)
        expected_params = {
            'follow': 'Alice,Bob',
            'track': 'foo,bar',
            'stall_warnings': 'true',
        }
        agent.add_expected_request('POST', uri, expected_params, stream)

        connected = Deferred()
        svc = client.stream_filter(
            lambda tweet: None, follow=['Alice', 'Bob'], track=['foo', 'bar'],
            stall_warnings=True)
        svc.set_connect_callback(connected.callback)
        svc.startService()
        connected_svc = yield connected
        self.assertIs(svc, connected_svc)
        yield svc.stopService()
        stream.finished()

    # TODO: Tests for stream_sample()
    # TODO: Tests for stream_firehose()

    @inlineCallbacks
    def test_userstream_user_with_user(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://userstream.twitter.com/1.1/user.json'
        stream = FakeResponse(None)
        agent.add_expected_request('GET', uri, {
            'stringify_friend_ids': 'true',
            'with': 'user',
        }, stream)

        connected = Deferred()
        tweets = []
        svc = client.userstream_user(tweets.append, with_='user')
        svc.set_connect_callback(connected.callback)
        svc.startService()
        connected_svc = yield connected
        self.assertIs(svc, connected_svc)
        self.assertEqual(tweets, [])

        stream.deliver_data(
            '{"friends_str": []}\r\n'
            '{"id_str": "1", "text": "Tweet 1", "user": {}}\r\n')
        self.assertEqual(tweets, [
            {"friends_str": []},
            {"id_str": "1", "text": "Tweet 1", "user": {}},
        ])
        stream.deliver_data(
            '{"id_str": "2", "text": "Tweet 2", "user": {}}\r\n')
        self.assertEqual(tweets, [
            {"friends_str": []},
            {"id_str": "1", "text": "Tweet 1", "user": {}},
            {"id_str": "2", "text": "Tweet 2", "user": {}},
        ])
        yield svc.stopService()
        stream.finished()

    @inlineCallbacks
    def test_userstream_user_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://userstream.twitter.com/1.1/user.json'
        stream = FakeResponse(None)
        agent.add_expected_request('GET', uri, {
            'stringify_friend_ids': 'true',
            'stall_warnings': 'true',
            'with': 'user',
            'replies': 'all',
        }, stream)

        connected = Deferred()
        svc = client.userstream_user(
            lambda tweet: None, stall_warnings=True, with_='user',
            replies='all')
        svc.set_connect_callback(connected.callback)
        svc.startService()
        connected_svc = yield connected
        self.assertIs(svc, connected_svc)
        yield svc.stopService()
        stream.finished()

    # Direct Messages

    @inlineCallbacks
    def test_direct_messages(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages.json'

        response_data = [{
            # Truncated dm data.
            "id": 1,
            "id_str": "1",
            "text": "hello",
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
        }]

        agent.add_expected_request(
            'GET', uri, {}, self._resp_json(response_data))

        resp = yield client.direct_messages()
        self.assertEqual(resp, response_data)

    @inlineCallbacks
    def test_direct_messages_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages.json'

        expected_params = {
            'count': '10',
            'since_id': '123',
            'max_id': '321',
            'skip_status': 'false',
            'include_entities': 'false',
        }

        response_data = [{
            # Truncated dm data.
            "id": 1,
            "id_str": "1",
            "text": "hello",
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
        }]

        agent.add_expected_request(
            'GET', uri, expected_params, self._resp_json(response_data))

        resp = yield client.direct_messages(
            count=10, since_id='123', max_id='321', skip_status=False,
            include_entities=False)
        self.assertEqual(resp, response_data)

    @inlineCallbacks
    def test_direct_messages_sent(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/sent.json'

        response_data = [{
            # Truncated dm data.
            "id": 1,
            "id_str": "1",
            "text": "hello",
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
        }]

        agent.add_expected_request(
            'GET', uri, {}, self._resp_json(response_data))

        resp = yield client.direct_messages_sent()
        self.assertEqual(resp, response_data)

    @inlineCallbacks
    def test_direct_messages_sent_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/sent.json'

        expected_params = {
            'count': '10',
            'since_id': '123',
            'max_id': '321',
            'page': '2',
            'include_entities': 'false',
        }

        response_data = [{
            # Truncated dm data.
            "id": 1,
            "id_str": "1",
            "text": "hello",
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
        }]

        agent.add_expected_request(
            'GET', uri, expected_params, self._resp_json(response_data))

        resp = yield client.direct_messages_sent(
            count=10, since_id='123', max_id='321', page=2,
            include_entities=False)
        self.assertEqual(resp, response_data)

    @inlineCallbacks
    def test_direct_messages_show(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/show.json'

        response_data = [{
            # Truncated dm data.
            "id": 1,
            "id_str": "1",
            "text": "hello",
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
        }]

        agent.add_expected_request(
            'GET', uri, {'id': '1'}, self._resp_json(response_data))

        resp = yield client.direct_messages_show('1')
        self.assertEqual(resp, response_data[0])

    def test_direct_messages_show_not_found(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/show.json'

        err_dict = {
            "errors": [{
                "message": "Sorry, that page does not exist",
                "code": 34,
            }]
        }

        agent.add_expected_request(
            'GET', uri, {'id': '1'}, self._resp_json(err_dict, 404))

        d = client.direct_messages_show(1)
        d.addErrback(lambda f: f.value)
        err = yield d

        code, _phrase, body = err.args
        self.assertEqual((404, err_dict), (code, json.loads(body)))

    def test_direct_messages_show_forbidden(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/show.json'

        err_dict = {
            "errors": [{
                "message": "There was an error sending your message: .",
                "code": 151
            }]
        }

        agent.add_expected_request(
            'GET', uri, {'id': '1'}, self._resp_json(err_dict, 403))

        d = client.direct_messages_show(1)
        d.addErrback(lambda f: f.value)
        err = yield d

        code, _phrase, body = err.args
        self.assertEqual((403, err_dict), (code, json.loads(body)))

    @inlineCallbacks
    def test_direct_messages_destroy(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/destroy.json'

        response_data = {
            # Truncated dm data.
            "id": 1,
            "id_str": "1",
            "text": "hello",
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
        }

        agent.add_expected_request(
            'POST', uri, {'id': '1'}, self._resp_json(response_data))

        resp = yield client.direct_messages_destroy('1')
        self.assertEqual(resp, response_data)

    def test_direct_messages_destroy_all_params(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/destroy.json'

        expected_params = {
            'id': '1',
            'include_entities': 'false'
        }

        response_data = {
            # Truncated dm data.
            "id": 1,
            "id_str": "1",
            "text": "hello",
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
        }

        agent.add_expected_request(
            'POST', uri, expected_params, self._resp_json(response_data))

        resp = yield client.direct_messages_destroy(
            '1', include_entities=False)
        self.assertEqual(resp, response_data)

    def test_direct_messages_destroy_not_found(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/destroy.json'

        err_dict = {
            "errors": [{
                "message": "Sorry, that page does not exist",
                "code": 34,
            }]
        }

        agent.add_expected_request(
            'GET', uri, {'id': '1'}, self._resp_json(err_dict, 404))

        d = client.direct_messages_destroy(1)
        d.addErrback(lambda f: f.value)
        err = yield d

        code, _phrase, body = err.args
        self.assertEqual((404, err_dict), (code, json.loads(body)))

    def test_direct_messages_destroy_forbidden(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/destroy.json'

        err_dict = {
            "errors": [{
                "message": "There was an error sending your message: .",
                "code": 151
            }]
        }

        agent.add_expected_request(
            'GET', uri, {'id': '1'}, self._resp_json(err_dict, 403))

        d = client.direct_messages_destroy(1)
        d.addErrback(lambda f: f.value)
        err = yield d

        code, _phrase, body = err.args
        self.assertEqual((403, err_dict), (code, json.loads(body)))

    @inlineCallbacks
    def test_direct_messages_new_by_user_id(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/new.json'

        expected_params = {
            'text': 'hello',
            'user_id': '2'
        }

        response_data = {
            # Truncated dm data.
            "id": 1,
            "id_str": "1",
            "text": "hello",
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
        }

        agent.add_expected_request(
            'POST', uri, expected_params, self._resp_json(response_data))

        resp = yield client.direct_messages_new('hello', user_id='2')
        self.assertEqual(resp, response_data)

    @inlineCallbacks
    def test_direct_messages_new_by_screen_name(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/new.json'

        expected_params = {
            'text': 'hello',
            'screen_name': 'fakeuser2'
        }

        response_data = {
            # Truncated dm data.
            "id": 1,
            "id_str": "1",
            "text": "hello",
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
        }

        agent.add_expected_request(
            'POST', uri, expected_params, self._resp_json(response_data))

        resp = yield client.direct_messages_new(
            'hello', screen_name='fakeuser2')
        self.assertEqual(resp, response_data)

    def test_direct_messages_new_bad_request(self):
        agent, client = self._agent_and_TwitterClient()
        uri = 'https://api.twitter.com/1.1/direct_messages/new.json'

        err_dict = {
            "errors": [{
                "message": (
                    "Recipient (user, screen name, or id) "
                    "parameter is missing."),
                "code": 38
            }]
        }

        agent.add_expected_request(
            'GET', uri, {'id': '1'}, self._resp_json(err_dict, 400))

        d = client.direct_messages_new('hello')
        d.addErrback(lambda f: f.value)
        err = yield d

        code, _phrase, body = err.args
        self.assertEqual((403, err_dict), (code, json.loads(body)))

    # Friends & Followers

    # TODO: Tests for friendships_no_retweets_ids()
    # TODO: Tests for friends_ids()
    # TODO: Tests for followers_ids()
    # TODO: Tests for friendships_lookup()
    # TODO: Tests for friendships_incoming()
    # TODO: Tests for friendships_outgoing()
    # TODO: Tests for friendships_create()
    # TODO: Tests for friendships_destroy()
    # TODO: Tests for friendships_update()
    # TODO: Tests for friendships_show()
    # TODO: Tests for friends_list()
    # TODO: Tests for followers_list()

    # Users

    # TODO: Tests for account_settings()
    # TODO: Tests for account_verify_credentials()
    # TODO: Tests for account_settings()
    # TODO: Tests for account_update_delivery_device()
    # TODO: Tests for account_update_profile()
    # TODO: Tests for account_update_profile_background_image()
    # TODO: Tests for account_update_profile_colors()
    # TODO: Tests for account_update_profile_image()
    # TODO: Tests for blocks_list()
    # TODO: Tests for blocks_ids()
    # TODO: Tests for blocks_create()
    # TODO: Tests for blocks_destroy()
    # TODO: Tests for users_lookup()
    # TODO: Tests for users_show()
    # TODO: Tests for users_search()
    # TODO: Tests for users_contributees()
    # TODO: Tests for users_contributors()
    # TODO: Tests for account_remove_profile_banner()
    # TODO: Tests for account_update_profile_banner()
    # TODO: Tests for users/profile_banner()

    # Suggested Users

    # TODO: Tests for users_suggestions()
    # TODO: Tests for users_suggestions()
    # TODO: Tests for users_suggestions_members()

    # Favorites

    # TODO: Tests for favorites_list()
    # TODO: Tests for favorites_destroy()
    # TODO: Tests for favorites_create()

    # Lists

    # TODO: Tests for lists_list()
    # TODO: Tests for lists_statuses()
    # TODO: Tests for lists_members_destroy()
    # TODO: Tests for lists_memberships()
    # TODO: Tests for lists_subscribers()
    # TODO: Tests for lists_subscribers/create()
    # TODO: Tests for lists_subscribers/show()
    # TODO: Tests for lists_subscribers/destroy()
    # TODO: Tests for lists_members_create_all()
    # TODO: Tests for lists_members_show()
    # TODO: Tests for lists_members()
    # TODO: Tests for lists_members_create()
    # TODO: Tests for lists_destroy()
    # TODO: Tests for lists_update()
    # TODO: Tests for lists_create()
    # TODO: Tests for lists_show()
    # TODO: Tests for lists_subscriptions()
    # TODO: Tests for lists_members_destroy_all()
    # TODO: Tests for lists_ownerships()

    # Saved Searches

    # TODO: Tests for saved_searches_list()
    # TODO: Tests for saved_searches_show()
    # TODO: Tests for saved_searches_create()
    # TODO: Tests for saved_searches_destroy()

    # Places & Geo

    # TODO: Tests for geo_id()
    # TODO: Tests for geo_reverse_geocode()
    # TODO: Tests for geo_search()
    # TODO: Tests for geo_similar_places()
    # TODO: Tests for geo_place()

    # Trends

    # TODO: Tests for trends_place()
    # TODO: Tests for trends_available()
    # TODO: Tests for trends_closest()

    # Spam Reporting

    # TODO: Tests for users_report_spam()

    # Help

    # TODO: Tests for help_configuration()
    # TODO: Tests for help_languages()
    # TODO: Tests for help_privacy()
    # TODO: Tests for help_tos()
    # TODO: Tests for application_rate_limit_status()



from datetime import datetime
from inspect import getmembers
import json
import re
from urlparse import urlparse, parse_qsl

from twisted.internet.defer import maybeDeferred

from txtwitter.error import TwitterAPIError
from txtwitter.tests.fake_agent import FakeResponse
from txtwitter.twitter import (
    TWITTER_API_URL, TWITTER_STREAM_URL, TWITTER_USERSTREAM_URL,
    TwitterClient)


USER_MENTION_RE = re.compile(r'@[a-zA-Z0-9_]+')


def mention_from_match(twitter_data, match):
    user = twitter_data.get_user_by_screen_name(match.group(0)[1:])

    if user is None:
        return None

    return {
        'id_str': user.id_str,
        'id': int(user.id_str),
        'indices': list(match.span(0)),
        'screen_name': user.screen_name,
        'name': user.name,
    }


def extract_user_mentions(twitter_data, text):
    mentions = []
    for match in USER_MENTION_RE.finditer(text):
        mention = mention_from_match(twitter_data, match)
        if mention is not None:
            mentions.append(mention)
    return mentions


def extract_entities(twitter_data, text):
    return {
        'user_mentions': extract_user_mentions(twitter_data, text),
        # TODO: More entities
    }


class FakeStream(object):
    def __init__(self):
        self.resp = FakeResponse(None)
        self._message_types = {}

    def add_message_type(self, message_type, predicate):
        self._message_types[message_type] = predicate

    def accepts(self, message_type, data):
        predicate = self._message_types.get(message_type)
        return predicate is not None and predicate(data)

    def deliver(self, data):
        self.resp.deliver_data(json.dumps(data))
        self.resp.deliver_data('\r\n')


class FakeTweet(object):
    def __init__(self, id_str, text, user_id_str, reply_to=None, **kw):
        self.id_str = id_str
        self.text = text
        self.user_id_str = user_id_str
        self.reply_to = reply_to
        self.created_at = kw.pop('created_at', datetime.utcnow())
        self.kw = kw

    def __cmp__(self, other):
        return cmp(int(self.id_str), int(other.id_str))

    def get_user(self, twitter_data):
        return twitter_data.get_user(self.user_id_str)

    def _get_reply_to_status_details(self, twitter_data):
        if self.reply_to is None:
            return {}

        reply_to_tweet = twitter_data.get_tweet(self.reply_to)

        return {
            'in_reply_to_status_id': int(reply_to_tweet.id_str),
            'in_reply_to_status_id_str': reply_to_tweet.id_str,
        }

    def _get_reply_to_user_details(self, twitter_data):
        def details(id_str, screen_name):
            return {
                'in_reply_to_user_id': int(id_str),
                'in_reply_to_user_id_str': id_str,
                'in_reply_to_screen_name': screen_name,
            }

        if self.reply_to is not None:
            reply_to_tweet = twitter_data.get_tweet(self.reply_to)
            reply_to_user = reply_to_tweet.get_user(twitter_data)
            return details(reply_to_user.id_str, reply_to_user.screen_name)

        match = USER_MENTION_RE.match(self.text)
        if match is not None:
            mention = mention_from_match(twitter_data, match)
            return details(mention['id_str'], mention['screen_name'])

        return {}

    def to_dict(self, twitter_data, trim_user=None, include_my_retweet=None,
                include_entities=None, contributor_details=None):
        if trim_user is None:
            trim_user = False
        if include_my_retweet is None:
            include_my_retweet = False
        if include_entities is None:
            include_entities = True
        if contributor_details is None:
            contributor_details = False
        else:
            raise NotImplementedError("contributer_details param")

        tweet_dict = {
            'id_str': self.id_str,
            'created_at': str(self.created_at),
            'text': self.text,
            # Defaults
            'favorite_count': 0,
            'filter_level': 'medium',
            'retweet_count': 0,
            'source': 'web',
        }
        # Calculated values
        tweet_dict.update({
            'id': int(tweet_dict['id_str']),
            'favorited': tweet_dict['favorite_count'] > 0,
            'retweeted': tweet_dict['retweet_count'] > 0,
        })
        if trim_user:
            tweet_dict['user'] = {
                'id_str': self.user_id_str,
                'id': int(self.user_id_str),
            }
        else:
            user = self.get_user(twitter_data)
            tweet_dict['user'] = user.to_dict(twitter_data)
        tweet_dict.update(self._get_reply_to_status_details(twitter_data))
        tweet_dict.update(self._get_reply_to_user_details(twitter_data))
        tweet_dict['entities'] = extract_entities(twitter_data, self.text)
        # Provided keyword args can override any of the above
        tweet_dict.update(self.kw)

        if not include_entities:
            tweet_dict.pop('entities')
        if include_my_retweet:
            raise NotImplementedError()
        return tweet_dict


class FakeDM(object):
    def __init__(self, id_str, text, sender_id_str, recipient_id_str, **kw):
        self.id_str = id_str
        self.text = text
        self.sender_id_str = sender_id_str
        self.recipient_id_str = recipient_id_str
        self.created_at = kw.pop('created_at', datetime.utcnow())
        self.kw = kw

    def __cmp__(self, other):
        return cmp(int(self.id_str), int(other.id_str))

    def _get_sender_details(self, twitter_data):
        sender = twitter_data.get_user(self.sender_id_str)
        return {
            'sender': sender.to_dict(twitter_data),
            'sender_id': int(self.sender_id_str),
            'sender_id_str': self.sender_id_str,
            'sender_screen_name': sender.screen_name
        }

    def _get_recipient_details(self, twitter_data):
        recipient = twitter_data.get_user(self.recipient_id_str)
        return {
            'recipient': recipient.to_dict(twitter_data),
            'recipient_id': int(self.recipient_id_str),
            'recipient_id_str': self.recipient_id_str,
            'recipient_screen_name': recipient.screen_name
        }

    def to_dict(self, twitter_data, skip_status=None,
                include_entities=None, **kw):
        if include_entities is None:
            include_entities = True
        if skip_status is not None:
            raise NotImplementedError("skip_status param")

        dm_dict = {
            'id': int(self.id_str),
            'id_str': self.id_str,
            'created_at': str(self.created_at),
            'text': self.text,
        }

        dm_dict.update(self._get_sender_details(twitter_data))
        dm_dict.update(self._get_recipient_details(twitter_data))
        dm_dict['entities'] = extract_entities(twitter_data, self.text)

        # Provided keyword args can override any of the above
        dm_dict.update(self.kw)

        if not include_entities:
            dm_dict.pop('entities')

        return dm_dict


class FakeUser(object):
    def __init__(self, id_str, screen_name, name, **kw):
        self.id_str = id_str
        self.screen_name = screen_name
        self.name = name
        self.created_at = kw.pop('created_at', datetime.utcnow())
        self.kw = kw

    def to_dict(self, twitter_data):
        user_dict = {
            'id_str': self.id_str,
            'screen_name': self.screen_name,
            'name': self.name,
            'created_at': str(self.created_at),
            # Defaults
        }
        # Calculated values
        user_dict.update({
            'id': int(user_dict['id_str']),
        })
        # Provided keyword args can override any of the above
        user_dict.update(self.kw)

        return user_dict


class FakeTwitterData(object):
    def __init__(self):
        self.users = {}
        self.dms = {}
        self.tweets = {}
        self.streams = {}
        self._next_dm_id = 1000
        self._next_tweet_id = 1000
        self._next_user_id = 1000

    @property
    def next_tweet_id(self):
        return str(self._next_tweet_id)

    @property
    def next_dm_id(self):
        return str(self._next_dm_id)

    @property
    def next_user_id(self):
        return str(self._next_user_id)

    def broadcast_tweet(self, tweet):
        for stream in self.streams.itervalues():
            if stream.accepts('tweet', tweet):
                stream.deliver(tweet.to_dict(self))

    def broadcast_dm(self, dm):
        for stream in self.streams.itervalues():
            if stream.accepts('dm', dm):
                stream.deliver({'direct_message': dm.to_dict(self)})

    def new_stream(self):
        stream = FakeStream()

        def finished_callback(r):
            self.remove_stream(stream.resp)

        stream.resp.finished_callback = finished_callback
        self.streams[stream.resp] = stream
        return stream

    def remove_stream(self, resp):
        self.streams.pop(resp, None)

    def get_tweet(self, id_str):
        return self.tweets.get(id_str)

    def get_dm(self, id_str):
        return self.dms.get(id_str)

    def get_user(self, id_str):
        return self.users.get(id_str)

    def add_tweet(self, *args, **kw):
        tweet = FakeTweet(*args, **kw)
        self.tweets[tweet.id_str] = tweet
        self.broadcast_tweet(tweet)
        return tweet

    def add_dm(self, *args, **kw):
        dm = FakeDM(*args, **kw)
        self.dms[dm.id_str] = dm
        self.broadcast_dm(dm)
        return dm

    def add_user(self, *args, **kw):
        user = FakeUser(*args, **kw)
        self.users[user.id_str] = user
        return user

    def del_tweet(self, id_str):
        self.tweets.pop(id_str)

    def del_dm(self, id_str):
        self.dms.pop(id_str)

    def del_user(self, id_str):
        self.users.pop(id_str)

    def new_tweet(self, text, user_id_str, *args, **kw):
        tweet = self.add_tweet(
            self.next_tweet_id, text, user_id_str, *args, **kw)
        self._next_tweet_id += 10
        return tweet

    def new_dm(self, text, sender_id_str, recipient_id_str, *args, **kw):
        dm = self.add_dm(
            self.next_dm_id, text, sender_id_str, recipient_id_str,
            *args, **kw)

        self._next_dm_id += 10
        return dm

    def new_user(self, screen_name, name, *args, **kw):
        user = self.add_user(self.next_user_id, screen_name, name, *args, **kw)
        self._next_user_id += 10
        return user

    def get_user_by_screen_name(self, screen_name):
        for user in self.users.itervalues():
            if user.screen_name == screen_name:
                return user

    def iter_tweets_from(self, user_id_str):
        for tweet in self.tweets.itervalues():
            if tweet.user_id_str == user_id_str:
                yield tweet

    def iter_tweets_mentioning(self, user_id_str):
        user = self.get_user(user_id_str)
        mention = '@%s' % (user.screen_name,)
        for tweet in self.tweets.itervalues():
            if mention in tweet.text:
                yield tweet

    def to_dicts(self, *objects, **kw):
        return [obj.to_dict(self, **kw) for obj in objects]


class FakeTwitterClient(TwitterClient):
    def __init__(self, fake_twitter, user_id_str,
                 api_url=TWITTER_API_URL, stream_url=TWITTER_STREAM_URL,
                 userstream_url=TWITTER_USERSTREAM_URL):
        self._fake_twitter = fake_twitter
        self._fake_twitter_user_id_str = user_id_str
        self._api_url_base = api_url
        self._stream_url_base = stream_url
        self._userstream_url_base = userstream_url

    def _make_request(self, method, uri, body_parameters=None):
        return self._fake_twitter.dispatch(
            self._fake_twitter_user_id_str, method, uri, body_parameters)

    def _parse_response(self, response):
        return response


class FakeTwitter(object):
    def __init__(self, api_url=TWITTER_API_URL, stream_url=TWITTER_STREAM_URL,
                 userstream_url=TWITTER_USERSTREAM_URL):
        self.urls = {
            'api': api_url,
            'stream': stream_url,
            'userstream': userstream_url,
        }
        self.twitter_data = FakeTwitterData()

    def __getattr__(self, name):
        return getattr(self.twitter_data, name)

    def get_client(self, user_id_str=None):
        return FakeTwitterClient(
            self, user_id_str, api_url=self.urls['api'],
            stream_url=self.urls['stream'],
            userstream_url=self.urls['userstream'])

    def get_api_method(self, user, uri):
        uri = uri.split('?')[0]

        user_view = FakeTwitterAPI(self.twitter_data, user)
        for name, method in getmembers(user_view):
            if not getattr(method, 'is_api', False):
                continue
            method_url_base = self.urls[method.api_host_prefix]
            if uri == method_url_base + method.api_path:
                return method

        raise ValueError("Unexpected URI: %r" % (uri,))

    def dispatch(self, user, method, uri, body_parameters):
        """
        Dispatch a fake request to the appropriate place.

        We don't actually care about the method here, since that's validated
        elsewhere and Twitter's API doesn't have different behaviour for the
        same URI depending on the method.

        """
        params = {}
        if body_parameters is not None:
            params.update(body_parameters)
        else:
            params.update(dict(parse_qsl(urlparse(uri).query)))

        method = self.get_api_method(user, uri)
        return maybeDeferred(method, **params)


def fake_api(path, host_prefix='api'):
    def deco(func):
        func.is_api = True
        func.api_path = path
        func.api_host_prefix = host_prefix
        return func
    return deco


class FakeTwitterAPI(object):
    def __init__(self, twitter_data, user_id_str):
        self._twitter_data = twitter_data
        self._user_id_str = user_id_str

    def _404(self):
        raise TwitterAPIError(404, "Not Found", json.dumps({
            "errors": [
                {"message": "Sorry, that page does not exist", "code": 34},
            ]}))

    def _tweet_or_404(self, id_str):
        tweet = self._twitter_data.get_tweet(id_str)
        if tweet is None:
            self._404()
        return tweet

    def _user_or_404(self, id_str):
        user = self._twitter_data.get_user(id_str)
        if user is None:
            self._404()
        return user

    def _dm_or_404(self, id_str):
        dm = self._twitter_data.get_dm(id_str)
        if dm is None:
            self._404()
        return dm

    # Timelines

    def _filter_timeline(self, tweets_iter, count, since_id, max_id):
        tweets = []
        for tweet in tweets_iter:
            tweet_id = int(tweet.id_str)
            if since_id is not None and tweet_id <= int(since_id):
                continue
            if max_id is not None and tweet_id > int(max_id):
                continue
            tweets.append(tweet)
        if count is None:
            count = 20
        if count > 200:
            count = 200
        return sorted(tweets, reverse=True)[:count]

    @fake_api('statuses/mentions_timeline.json')
    def statuses_mentions_timeline(self, count=None, since_id=None,
                                   max_id=None, trim_user=None,
                                   contributor_details=None,
                                   include_entities=None):
        tweets = self._filter_timeline(
            self._twitter_data.iter_tweets_mentioning(self._user_id_str),
            count, since_id, max_id)
        return [
            tweet.to_dict(
                self._twitter_data, trim_user=trim_user,
                contributor_details=contributor_details,
                include_entities=include_entities)
            for tweet in tweets]

    @fake_api('statuses/user_timeline.json')
    def statuses_user_timeline(self, user_id=None, screen_name=None,
                               since_id=None, count=None, max_id=None,
                               trim_user=None, exclude_replies=None,
                               contributor_details=None,
                               include_rts=None):
        if user_id is None and screen_name is not None:
            user = self._twitter_data.get_user_by_screen_name(screen_name)
            user_id = user.id_str
            screen_name = None
        if user_id is None or screen_name is not None:
            raise NotImplementedError("user_id/screen_name params")

        tweets = self._filter_timeline(
            self._twitter_data.iter_tweets_from(user_id), count, since_id,
            max_id)
        if exclude_replies:
            tweets = [tweet for tweet in tweets if tweet.reply_to is None]
        if include_rts is not None:
            raise NotImplementedError("exclude_rts param")
        return [
            tweet.to_dict(
                self._twitter_data, trim_user=trim_user,
                contributor_details=contributor_details)
            for tweet in tweets]

    @fake_api('statuses/home_timeline.json')
    def statuses_home_timeline(self, count=None, since_id=None, max_id=None,
                               trim_user=None, exclude_replies=None,
                               contributor_details=None,
                               include_entities=None):
        raise NotImplementedError()

    # TODO: Implement statuses_retweets_of_me()

    # Tweets

    @fake_api('statuses/retweets.json')
    def statuses_retweets(self, id, count=None, trim_user=None):
        raise NotImplementedError()

    @fake_api('statuses/show.json')
    def statuses_show(self, id, trim_user=None, include_my_retweet=None,
                      include_entities=None):
        return self._tweet_or_404(id).to_dict(
            self._twitter_data, trim_user=trim_user,
            include_my_retweet=include_my_retweet,
            include_entities=include_entities)

    @fake_api('statuses/destroy.json')
    def statuses_destroy(self, id, trim_user=None):
        tweet = self._tweet_or_404(id)
        # TODO: Better error
        if tweet.user_id_str != self._user_id_str:
            raise NotImplementedError()
        tweet_dict = tweet.to_dict(self._twitter_data, trim_user=trim_user)
        self._twitter_data.del_tweet(id)
        return tweet_dict

    @fake_api('statuses/update.json')
    def statuses_update(self, status, in_reply_to_status_id=None, lat=None,
                        long=None, place_id=None, display_coordinates=None,
                        trim_user=None):
        if set([lat, long, place_id, display_coordinates]) != set([None]):
            raise NotImplementedError("Unsupported parameter")
        tweet = self._twitter_data.new_tweet(
            status, self._user_id_str, reply_to=in_reply_to_status_id)
        return tweet.to_dict(self._twitter_data, trim_user=trim_user)

    @fake_api('statuses/retweet.json')
    def statuses_retweet(self, id, trim_user=None):
        raise NotImplementedError()

    # TODO: Implement statuses_update_with_media()
    # TODO: Implement statuses_oembed()
    # TODO: Implement statuses_retweeters_ids()

    # Search

    # TODO: Implement search_tweets()

    # Streaming

    @fake_api('statuses/filter.json', 'stream')
    def stream_filter(self, follow=None, track=None, locations=None,
                      stall_warnings=None):
        track_res = []
        if track:
            for term in track.split(','):
                track_res.append(re.compile(r'\b%s\b' % (re.escape(term),)))

        follow = [] if follow is None else follow.split(',')

        def stream_filter_predicate(tweet):
            for user_id_str in follow:
                if tweet.user_id_str == user_id_str:
                    return True
            for track_re in track_res:
                if track_re.search(tweet.text):
                    return True
            return False

        stream = self._twitter_data.new_stream()
        stream.add_message_type('tweet', stream_filter_predicate)
        return stream.resp

    # TODO: Implement stream_sample()
    # TODO: Implement stream_firehose()

    @fake_api('user.json', 'userstream')
    def userstream_user(self, stringify_friend_ids, stall_warnings=None,
                        with_='followings', replies=None, **kw):
        with_ = kw.pop('with', with_)
        assert kw == {}
        user = self._twitter_data.get_user(self._user_id_str)
        mention_re = re.compile(r'@%s\b' % (user.screen_name,))

        if with_ != 'user':
            raise NotImplementedError("with != followings")

        def userstream_tweet_predicate(tweet):
            if tweet.user_id_str == self._user_id_str:
                return True
            if mention_re.search(tweet.text):
                return True
            if with_ == 'followings':
                pass
            return False

        def userstream_dm_predicate(dm):
            if dm.recipient_id_str == self._user_id_str:
                return True
            if dm.sender_id_str == self._user_id_str:
                return True
            return False

        stream = self._twitter_data.new_stream()
        stream.add_message_type('tweet', userstream_tweet_predicate)
        stream.add_message_type('dm', userstream_dm_predicate)

        # TODO: Proper friends.
        stream.deliver({'friends_str': []})

        return stream.resp

    # Direct Messages
    def _clamp_dms(self, dms, since_id=None, max_id=None, count=None):
        if since_id is not None:
            since_id = int(since_id)
            dms = [dm for dm in dms if int(dm.id_str) > since_id]
        if max_id is not None:
            max_id = int(max_id)
            dms = [dm for dm in dms if int(dm.id_str) <= max_id]

        dms = sorted(dms, reverse=True)
        count = 20 if count is None else min(count, 200)
        return dms[:count]

    @fake_api('direct_messages.json')
    def direct_messages(self, since_id=None, max_id=None, count=None,
                        include_entities=None, skip_status=None):
        dms = self._twitter_data.dms.values()
        dms = [dm for dm in dms if dm.recipient_id_str == self._user_id_str]
        dms = self._clamp_dms(dms, since_id, max_id, count)

        return self._twitter_data.to_dicts(
            *dms, include_entities=include_entities, skip_status=skip_status)

    @fake_api('direct_messages/sent.json')
    def direct_messages_sent(self, since_id=None, max_id=None, count=None,
                             include_entities=None, page=None):
        dms = self._twitter_data.dms.values()
        dms = [dm for dm in dms if dm.sender_id_str == self._user_id_str]
        dms = self._clamp_dms(dms, since_id, max_id, count)

        if page is None:
            page = 1

        page_end = page * 20
        dms = dms[page_end - 20:page_end]

        return self._twitter_data.to_dicts(
            *dms, include_entities=include_entities)

    @fake_api('direct_messages/show.json')
    def direct_messages_show(self, id):
        dm = self._dm_or_404(id)

        if (dm.recipient_id_str != self._user_id_str and
                dm.sender_id_str != self._user_id_str):
            # The actual response given to us by Twitter
            raise TwitterAPIError(403, "Forbidden", json.dumps({
                "errors": [{
                    "message": "There was an error sending your message: .",
                    "code": 151
                }]}))

        return self._twitter_data.to_dicts(dm)

    @fake_api('direct_messages/destroy.json')
    def direct_messages_destroy(self, id, include_entities=None):
        dm = self._dm_or_404(id)

        if (dm.recipient_id_str != self._user_id_str and
                dm.sender_id_str != self._user_id_str):
            # The actual response given to us by Twitter
            raise TwitterAPIError(403, "Forbidden", json.dumps({
                "errors": [{
                    "message": "There was an error sending your message: .",
                    "code": 151
                }]}))

        del self._twitter_data.dms[dm.id_str]
        return dm.to_dict(
            self._twitter_data, include_entities=include_entities)

    @fake_api('direct_messages/new.json')
    def direct_messages_new(self, text, user_id=None, screen_name=None):
        if user_id is None and screen_name is None:
            raise TwitterAPIError(400, "Bad Request", json.dumps({
                "errors": [{
                    "message": (
                        "Recipient (user, screen name, or id) "
                        "parameter is missing."),
                    "code": 38
                }]}))

        if user_id is None:
            user = self._twitter_data.get_user_by_screen_name(screen_name)
            user_id = user.id_str

        dm = self._twitter_data.new_dm(text, self._user_id_str, user_id)
        return dm.to_dict(self._twitter_data)

    # Friends & Followers

    # TODO: Implement friendships_no_retweets_ids()
    # TODO: Implement friends_ids()
    # TODO: Implement followers_ids()
    # TODO: Implement friendships_lookup()
    # TODO: Implement friendships_incoming()
    # TODO: Implement friendships_outgoing()
    # TODO: Implement friendships_create()
    # TODO: Implement friendships_destroy()
    # TODO: Implement friendships_update()
    # TODO: Implement friendships_show()
    # TODO: Implement friends_list()
    # TODO: Implement followers_list()

    # Users

    # TODO: Implement account_settings()
    # TODO: Implement account_verify_credentials()
    # TODO: Implement account_settings()
    # TODO: Implement account_update_delivery_device()
    # TODO: Implement account_update_profile()
    # TODO: Implement account_update_profile_background_image()
    # TODO: Implement account_update_profile_colors()
    # TODO: Implement account_update_profile_image()
    # TODO: Implement blocks_list()
    # TODO: Implement blocks_ids()
    # TODO: Implement blocks_create()
    # TODO: Implement blocks_destroy()
    # TODO: Implement users_lookup()
    # TODO: Implement users_show()
    # TODO: Implement users_search()
    # TODO: Implement users_contributees()
    # TODO: Implement users_contributors()
    # TODO: Implement account_remove_profile_banner()
    # TODO: Implement account_update_profile_banner()
    # TODO: Implement users/profile_banner()

    # Suggested Users

    # TODO: Implement users_suggestions()
    # TODO: Implement users_suggestions()
    # TODO: Implement users_suggestions_members()

    # Favorites

    # TODO: Implement favorites_list()
    # TODO: Implement favorites_destroy()
    # TODO: Implement favorites_create()

    # Lists

    # TODO: Implement lists_list()
    # TODO: Implement lists_statuses()
    # TODO: Implement lists_members_destroy()
    # TODO: Implement lists_memberships()
    # TODO: Implement lists_subscribers()
    # TODO: Implement lists_subscribers/create()
    # TODO: Implement lists_subscribers/show()
    # TODO: Implement lists_subscribers/destroy()
    # TODO: Implement lists_members_create_all()
    # TODO: Implement lists_members_show()
    # TODO: Implement lists_members()
    # TODO: Implement lists_members_create()
    # TODO: Implement lists_destroy()
    # TODO: Implement lists_update()
    # TODO: Implement lists_create()
    # TODO: Implement lists_show()
    # TODO: Implement lists_subscriptions()
    # TODO: Implement lists_members_destroy_all()
    # TODO: Implement lists_ownerships()

    # Saved Searches

    # TODO: Implement saved_searches_list()
    # TODO: Implement saved_searches_show()
    # TODO: Implement saved_searches_create()
    # TODO: Implement saved_searches_destroy()

    # Places & Geo

    # TODO: Implement geo_id()
    # TODO: Implement geo_reverse_geocode()
    # TODO: Implement geo_search()
    # TODO: Implement geo_similar_places()
    # TODO: Implement geo_place()

    # Trends

    # TODO: Implement trends_place()
    # TODO: Implement trends_available()
    # TODO: Implement trends_closest()

    # Spam Reporting

    # TODO: Implement users_report_spam()

    # OAuth

    # TODO: Decide whether any of these APIs should be implemented.

    # Help

    # TODO: Implement help_configuration()
    # TODO: Implement help_languages()
    # TODO: Implement help_privacy()
    # TODO: Implement help_tos()
    # TODO: Implement application_rate_limit_status()



from twisted.trial.unittest import TestCase


class TestTweetFunctions(TestCase):
    def setUp(self):
        from txtwitter import messagetools
        self.messagetools = messagetools

    def test_is_tweet(self):
        """
        is_tweet() should return `True` for a tweet message.
        """
        self.assertEqual(True, self.messagetools.is_tweet({
            'id_str': '12345',
            'text': 'This is a tweet.',
            'user': {},
        }))

    def test_is_tweet_nontweet(self):
        """
        is_tweet() should return `False` for a non-tweet message.
        """
        self.assertEqual(False, self.messagetools.is_tweet({
            'friends': [],
        }))

    def test_ensure_tweet(self):
        """
        ensure_tweet() should return the original message for a tweet message.
        """
        msg = {
            'id_str': '12345',
            'text': 'This is a tweet.',
            'user': {},
        }
        self.assertEqual(msg, self.messagetools.ensure_tweet(msg))

    def test_ensure_tweet_nontweet(self):
        """
        ensure_tweet() should raise `ValueError` for a non-tweet message.
        """
        msg = {'friends': []}
        self.assertRaises(ValueError, self.messagetools.ensure_tweet, msg)

    def test_tweet_id(self):
        """
        tweet_id() should return the `id_str` field of a tweet message.
        """
        msg = {
            'id_str': '12345',
            'text': 'This is a tweet.',
            'user': {},
        }
        self.assertEqual('12345', self.messagetools.tweet_id(msg))

    def test_tweet_id_nontweet(self):
        """
        tweet_id() should raise `ValueError` for a non-tweet message.
        """
        msg = {'friends': []}
        self.assertRaises(ValueError, self.messagetools.tweet_id, msg)

    def test_tweet_text(self):
        """
        tweet_text() should return the `text` field of a tweet message.
        """
        msg = {
            'id_str': '12345',
            'text': 'This is a tweet.',
            'user': {},
        }
        self.assertEqual('This is a tweet.', self.messagetools.tweet_text(msg))

    def test_tweet_text_nontweet(self):
        """
        tweet_text() should raise `ValueError` for a non-tweet message.
        """
        msg = {'friends': []}
        self.assertRaises(ValueError, self.messagetools.tweet_text, msg)

    def test_tweet_user_mentions(self):
        """
        tweet_user_mentions() should return the list of users mentioned in the
        tweet message.
        """
        user_mention = {
            'id_str': '123',
            'screen_name': 'fakeuser',
            'name': 'Fake User',
        }
        msg = {
            'entities': {'user_mentions': [user_mention]},
            'id_str': '12345',
            'text': 'This is a tweet mentioning @fakeuser.',
            'user': {},
        }
        self.assertEqual(
            [user_mention], self.messagetools.tweet_user_mentions(msg))

    def test_tweet_user_mentions_no_mentions(self):
        """
        tweet_user_mentions() should return an empty list if no users are
        mentioned in the tweet message.
        """
        msg = {
            'entities': {'user_mentions': []},
            'id_str': '12345',
            'text': 'This is a tweet mentioning @fakeuser.',
            'user': {},
        }
        self.assertEqual([], self.messagetools.tweet_user_mentions(msg))

    def test_tweet_user_mentions_nontweet(self):
        """
        tweet_user_mentions() should raise `ValueError` for a non-tweet
        message.
        """
        msg = {'friends': []}
        self.assertRaises(
            ValueError, self.messagetools.tweet_user_mentions, msg)

    def test_tweet_in_reply_to_id(self):
        """
        tweet_in_reply_to_id() should return the id of the tweet this tweet
        message is a reply to.
        """
        msg = {
            'id_str': '12345',
            'in_reply_to_status_id_str': '12344',
            'text': '@fakeuser This is a reply.',
            'user': {},
        }
        self.assertEqual('12344', self.messagetools.tweet_in_reply_to_id(msg))

    def test_tweet_in_reply_to_id_nonreply(self):
        """
        tweet_in_reply_to_id() should return `None` if this tweet message is
        not a reply.
        """
        msg = {
            'id_str': '12345',
            'in_reply_to_status_id_str': None,
            'text': '@fakeuser This is a reply.',
            'user': {},
        }
        self.assertEqual(None, self.messagetools.tweet_in_reply_to_id(msg))

    def test_tweet_in_reply_to_id_nontweet(self):
        """
        tweet_in_reply_to_id() should raise `ValueError` for a non-tweet
        message.
        """
        msg = {'friends': []}
        self.assertRaises(
            ValueError, self.messagetools.tweet_in_reply_to_id, msg)

    def test_tweet_in_reply_to_screen_name(self):
        """
        tweet_in_reply_to_screen_name() should return the screen name of the
        user who's tweet this tweet message is a reply to.
        """
        msg = {
            'id_str': '12345',
            'in_reply_to_status_id_str': '12344',
            'in_reply_to_screen_name': 'sam',
            'text': '@fakeuser This is a reply.',
            'user': {},
        }
        self.assertEqual(
            'sam', self.messagetools.tweet_in_reply_to_screen_name(msg))

    def test_tweet_in_reply_to_screen_name_nonreply(self):
        """
        tweet_in_reply_to_screen_name() should return `None` if this tweet
        message is not a reply.
        """
        msg = {
            'id_str': '12345',
            'in_reply_to_status_id_str': None,
            'in_reply_to_screen_name': None,
            'text': '@fakeuser This is a reply.',
            'user': {},
        }
        self.assertEqual(
            None, self.messagetools.tweet_in_reply_to_screen_name(msg))

    def test_tweet_in_reply_to_screen_name_nontweet(self):
        """
        tweet_in_reply_to_screen_name() should raise `ValueError` for a
        non-tweet message.
        """
        msg = {'friends': []}
        self.assertRaises(
            ValueError, self.messagetools.tweet_in_reply_to_screen_name, msg)

    def test_tweet_is_reply(self):
        """
        tweet_in_reply_to_id() should return `True` if this tweet message is a
        reply.
        """
        msg = {
            'id_str': '12345',
            'in_reply_to_status_id_str': '12344',
            'text': '@fakeuser This is a reply.',
            'user': {},
        }
        self.assertEqual(True, self.messagetools.tweet_is_reply(msg))

    def test_tweet_is_reply_nonreply(self):
        """
        tweet_is_reply() should return `False` if this tweet message is not a
        reply.
        """
        msg = {
            'id_str': '12345',
            'in_reply_to_status_id_str': None,
            'text': '@fakeuser This is a reply.',
            'user': {},
        }
        self.assertEqual(False, self.messagetools.tweet_is_reply(msg))

    def test_tweet_is_reply_nontweet(self):
        """
        tweet_is_reply() should raise `ValueError` for a non-tweet message.
        """
        msg = {'friends': []}
        self.assertRaises(ValueError, self.messagetools.tweet_is_reply, msg)

    def test_tweet_user(self):
        """
        tweet_user() should return the `user` field of a tweet message.
        """
        msg = {
            'id_str': '12345',
            'text': 'This is a tweet.',
            'user': {
                'id_str': '67890',
                'screen_name': 'luke',
            },
        }
        self.assertEqual(msg['user'], self.messagetools.tweet_user(msg))

    def test_tweet_user_nontweet(self):
        """
        tweet_user() should raise `ValueError` for a non-tweet message.
        """
        msg = {'friends': []}
        self.assertRaises(ValueError, self.messagetools.tweet_user, msg)

    def test_is_dm(self):
        """
        is_dm() should return `True` for a dm message.
        """
        self.assertEqual(True, self.messagetools.is_dm({
            "id": 1,
            "id_str": "1",
            'text': 'This is a dm.',
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "sender": {},
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
            "recipient": {},
        }))

    def test_is_dm_nondm(self):
        """
        is_dm() should return `False` for a non-dm message.
        """
        self.assertEqual(False, self.messagetools.is_dm({
            'friends': [],
        }))

    def test_ensure_dm(self):
        """
        ensure_dm() should return the original message for a dm message.
        """
        msg = {
            "id": 1,
            "id_str": "1",
            'text': 'This is a dm.',
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "sender": {},
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
            "recipient": {},
        }
        self.assertEqual(msg, self.messagetools.ensure_dm(msg))

    def test_ensure_dm_nondm(self):
        """
        ensure_dm() should raise `ValueError` for a non-dm message.
        """
        msg = {'friends': []}
        self.assertRaises(ValueError, self.messagetools.ensure_dm, msg)

    def test_dm_id(self):
        """
        dm_id() should return the `id_str` field of a dm message.
        """
        msg = {
            "id_str": "1",
            'text': 'This is a dm.',
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "sender": {},
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
            "recipient": {},
        }
        self.assertEqual('1', self.messagetools.dm_id(msg))

    def test_dm_id_nondm(self):
        """
        dm_id() should raise `ValueError` for a non-dm message.
        """
        msg = {'friends': []}
        self.assertRaises(ValueError, self.messagetools.dm_id, msg)

    def test_dm_text(self):
        """
        dm_text() should return the `text` field of a dm message.
        """
        msg = {
            "id_str": "1",
            'text': 'This is a dm.',
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "sender": {},
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
            "recipient": {},
        }
        self.assertEqual('This is a dm.', self.messagetools.dm_text(msg))

    def test_dm_text_nondm(self):
        """
        dm_text() should raise `ValueError` for a non-dm message.
        """
        msg = {'friends': []}
        self.assertRaises(ValueError, self.messagetools.dm_text, msg)

    def test_dm_user_mentions(self):
        """
        dm_user_mentions() should return the list of users mentioned in the
        dm message.
        """
        user_mention = {
            'id_str': '123',
            'screen_name': 'fakeuser2',
            'name': 'Fake User 2',
        }
        msg = {
            "id_str": "1",
            'text': 'This is a dm mentioning @fakeuser2.',
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "sender": {},
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
            "recipient": {},
            'entities': {'user_mentions': [user_mention]},
        }
        self.assertEqual(
            [user_mention], self.messagetools.dm_user_mentions(msg))

    def test_dm_user_mentions_no_mentions(self):
        """
        dm_user_mentions() should return an empty list if no users are
        mentioned in the dm message.
        """
        msg = {
            "id_str": "1",
            'text': 'This is a dm.',
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "sender": {},
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
            "recipient": {},
            'entities': {'user_mentions': []},
        }
        self.assertEqual([], self.messagetools.dm_user_mentions(msg))

    def test_dm_user_mentions_nondm(self):
        """
        dm_user_mentions() should raise `ValueError` for a non-dm message.
        """
        msg = {'friends': []}
        self.assertRaises(
            ValueError, self.messagetools.dm_user_mentions, msg)

    def test_dm_sender(self):
        """
        dm_sender() should return the `sender` field of a dm message.
        """
        msg = {
            "id_str": "1",
            'text': 'This is a dm mentioning @fakeuser2.',
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "sender": {
                "id": 1,
                "id_str": "1",
                "screen_name": "fakeuser",
            },
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
            "recipient": {},
        }
        self.assertEqual(msg['sender'], self.messagetools.dm_sender(msg))

    def test_dm_sender_nondm(self):
        """
        dm_sender() should raise `ValueError` for a non-dm message.
        """
        msg = {'friends': []}
        self.assertRaises(ValueError, self.messagetools.dm_sender, msg)

    def test_dm_recipient(self):
        """
        dm_recipient() should return the `recipient` field of a dm message.
        """
        msg = {
            "id_str": "1",
            'text': 'This is a dm.',
            "sender_id": 1,
            "sender_id_str": "1",
            "sender_screen_name": "fakeuser",
            "sender": {},
            "recipient_id": 2,
            "recipient_id_str": "2",
            "recipient_screen_name": "fakeuser2",
            "recipient": {
                "id": 2,
                "id_str": "2",
                "screen_name": "fakeuser2",
            },
        }
        self.assertEqual(msg['recipient'], self.messagetools.dm_recipient(msg))

    def test_dm_recipient_nondm(self):
        """
        dm_recipient() should raise `ValueError` for a non-dm message.
        """
        msg = {'friends': []}
        self.assertRaises(ValueError, self.messagetools.dm_recipient, msg)

    def test_is_user(self):
        """
        is_user() should return `True` for valid user data.
        """
        self.assertEqual(True, self.messagetools.is_user({
            'id_str': '12345',
            'screen_name': 'luke'
        }))

    def test_is_user_nontweet(self):
        """
        is_user() should return `False` for a non-tweet message.
        """
        self.assertEqual(False, self.messagetools.is_user({
            'friends': [],
        }))

    def test_ensure_user(self):
        """
        ensure_user() should return the original user data.
        """
        user = {
            'id_str': '12345',
            'screen_name': 'luke'
        }
        self.assertEqual(user, self.messagetools.ensure_user(user))

    def test_ensure_user_nonuser(self):
        """
        ensure_user() should raise `ValueError` for invalid user data.
        """
        self.assertRaises(ValueError, self.messagetools.ensure_user, {})

    def test_user_id(self):
        """
        user_id() should return the `id_str` field of a user.
        """
        user = {
            'id_str': '12345',
            'screen_name': 'luke'
        }
        self.assertEqual('12345', self.messagetools.user_id(user))

    def test_user_id_nonuser(self):
        """
        user_id() should raise `ValueError` for a non-tweet message.
        """
        self.assertRaises(ValueError, self.messagetools.user_id, {})

    def test_user_screen_name(self):
        """
        user_screen_name() should return the `screen_name` field of a user.
        """
        user = {
            'id_str': '12345',
            'screen_name': 'luke'
        }
        self.assertEqual('luke', self.messagetools.user_screen_name(user))

    def test_user_screen_name_nonuser(self):
        """
        user_screen_name() should raise `ValueError` for a non-tweet message.
        """
        self.assertRaises(ValueError, self.messagetools.user_screen_name, {})



"""
A collection of utilities for working with Twitter API messages.
"""


def is_tweet(message):
    return set(['id_str', 'text', 'user']).issubset(set(message.keys()))


def ensure_tweet(message):
    if not is_tweet(message):
        raise ValueError("Message is not a tweet: %r" % (message,))
    return message


def tweet_text(message):
    return ensure_tweet(message)['text']


def tweet_user_mentions(message):
    return ensure_tweet(message)['entities'].get('user_mentions', [])


def tweet_id(message):
    return ensure_tweet(message)['id_str']


def tweet_in_reply_to_id(message):
    return ensure_tweet(message).get('in_reply_to_status_id_str', None)


def tweet_in_reply_to_screen_name(message):
    return ensure_tweet(message).get('in_reply_to_screen_name', None)


def tweet_is_reply(message):
    return tweet_in_reply_to_id(message) is not None


def tweet_user(message):
    return ensure_tweet(message)['user']


def is_dm(message):
    fields = ['id_str', 'text', 'sender', 'recipient']
    return set(fields).issubset(set(message.keys()))


def ensure_dm(message):
    if not is_dm(message):
        raise ValueError("Message is not a direct message: %r" % (message,))
    return message


def dm_text(message):
    return ensure_dm(message)['text']


def dm_user_mentions(message):
    return ensure_dm(message)['entities'].get('user_mentions', [])


def dm_id(message):
    return ensure_dm(message)['id_str']


def dm_sender(message):
    return ensure_dm(message)['sender']


def dm_recipient(message):
    return ensure_dm(message)['recipient']


def is_user(user):
    return set(['id_str', 'screen_name']).issubset(set(user.keys()))


def ensure_user(user):
    if not is_user(user):
        raise ValueError("Data is not user data: %r" % (user,))
    return user


def user_id(user):
    return ensure_user(user).get('id_str', None)


def user_screen_name(user):
    return ensure_user(user).get('screen_name', None)




from datetime import datetime

from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.web.client import ResponseDone
from twisted.trial.unittest import TestCase


def skipTest(reason):
    def deco(func):
        func.skip = reason
        return func
    return deco


class TestTwitterTransportRealIntertrons(TestCase):
    timeout = 15

    def setUp(self):
        from txtwitter import twitter
        self.TwitterClient = twitter.TwitterClient

        self.config = {
            'consumer_key': 'MBRx2eTDZu1Eu2VHiGSLw',
            'consumer_secret': '9jxvZdCwLyXSvrrWibgzjbtOdRjLWdl3TobUrvpo9Js',
            'token_key':
                '2258763672-JXsurIhVkdsuZdikIq0xiUPlPSj4X65zjP257Nx',
            'token_secret':
                'jcy7U9PuBmWuKtGzdxOyE2aw05y6SBMY2GuAxb4FYmIgf',
        }

    @skipTest("Don't hit the real API for now.")
    @inlineCallbacks
    def test_statuses_show(self):
        client = self.TwitterClient(**self.config)
        resp = yield client.statuses_show("415184968248487936", trim_user=True)
        print "\n-----"
        print resp
        print "-----"

    @skipTest("Don't hit the real API for now.")
    @inlineCallbacks
    def test_show_bad_id(self):
        client = self.TwitterClient(**self.config)
        d = client.show("badtweet")
        d.addErrback(lambda f: f.value)
        err = yield d
        print "\n-----"
        print repr(err)
        print "-----"

    @skipTest("Don't hit the real API for now.")
    @inlineCallbacks
    def test_statuses_update(self):
        client = self.TwitterClient(**self.config)
        resp = yield client.statuses_update(
            "new test: %s" % (datetime.utcnow(),))
        print "\n-----"
        print resp
        print "-----"

    @skipTest("Don't hit the real API for now.")
    @inlineCallbacks
    def test_statuses_update_mentions(self):
        client = self.TwitterClient(**self.config)
        resp = yield client.statuses_update(
            ".@jerithtest @E5BB1EF8_6EB9 @jerithtest2 %s" % (
                datetime.utcnow(),))
        print "\n-----"
        print resp
        print "-----"

    @skipTest("Don't hit the real API for now.")
    @inlineCallbacks
    def test_stream_filter(self):

        class LimitedDelegate(object):
            def __init__(self, max_count):
                self.count = 0
                self.max_count = max_count

            def __call__(self, tweet):
                self.count += 1
                print "\n-- v %s v ---" % self.count
                print tweet
                import json
                print "=== %s ===" % len(json.dumps(tweet))
                print "-- ^ %s ^ ---" % self.count
                if self.count >= self.max_count:
                    self.svc.stopService()

        client = self.TwitterClient(**self.config)
        delegate = LimitedDelegate(5)
        svc = client.stream_filter(delegate, track=["twitter"])
        d = Deferred()
        svc.set_disconnect_callback(lambda s, r: d.callback(r))
        delegate.svc = svc
        yield d.addErrback(lambda f: f.trap(ResponseDone))

    @skipTest("Don't hit the real API for now.")
    @inlineCallbacks
    def test_userstream_user(self):

        class LimitedDelegate(object):
            def __init__(self, max_count):
                self.count = 0
                self.max_count = max_count

            def __call__(self, tweet):
                self.count += 1
                print "\n-- v %s v ---" % self.count
                print tweet
                import json
                print "=== %s ===" % len(json.dumps(tweet))
                print "-- ^ %s ^ ---" % self.count
                if self.count >= self.max_count:
                    self.svc.stopService()

        client = self.TwitterClient(**self.config)
        delegate = LimitedDelegate(5)
        svc = client.userstream_user(delegate, with_='user')
        d = Deferred()
        svc.set_disconnect_callback(lambda s, r: d.callback(r))
        delegate.svc = svc
        yield d.addErrback(lambda f: f.trap(ResponseDone))



from twisted.internet.defer import Deferred
from twisted.internet.task import Clock
from twisted.python.failure import Failure
from twisted.trial.unittest import TestCase
from twisted.web.client import ResponseDone

from txtwitter.tests.fake_agent import FakeResponse


def from_streamservice(name):
    @property
    def prop(self):
        from txtwitter import streamservice
        return getattr(streamservice, name)
    return prop


class TestTwitterClient(TestCase):
    _TwitterStreamService = from_streamservice('TwitterStreamService')

    def test_set_connect_callback(self):
        """
        set_connect_callback() should set the service's connect callback.
        """
        svc = self._TwitterStreamService(None, None)
        self.assertEqual(svc.connect_callback, None)
        svc.set_connect_callback('foo')
        self.assertEqual(svc.connect_callback, 'foo')

    def test_connect_callback(self):
        """
        The connect callback should be called on a successful connection.
        """
        d = Deferred()
        called = []
        svc = self._TwitterStreamService(lambda: d, None)
        svc.set_connect_callback(lambda s: called.append(s))
        svc.startService()
        self.assertEqual(called, [])
        d.callback(FakeResponse(None))
        self.assertEqual(called, [svc])

    def test_connect_callback_None(self):
        """
        The connect callback should not be called if it is unset.

        It's hard to assert that something /doesn't/ happen, so we make sure we
        see something else that happens at connect time and assert that no
        errors were logged.
        """
        d = Deferred()
        svc = self._TwitterStreamService(lambda: d, None)
        svc.startService()
        self.assertEqual(None, svc.connect_callback)
        self.assertEqual(None, svc._stream_response)
        d.callback(FakeResponse(None))
        self.assertNotEqual(None, svc._stream_response)
        self.assertEqual([], self.flushLoggedErrors())

    def test_set_disconnect_callback(self):
        """
        set_disconnect_callback() should set the service's disconnect callback.
        """
        svc = self._TwitterStreamService(None, None)
        self.assertEqual(svc.disconnect_callback, None)
        svc.set_disconnect_callback('foo')
        self.assertEqual(svc.disconnect_callback, 'foo')

    def test_disconnect_callback(self):
        """
        The disconnect callback should be called on disconnection.
        """
        called = []
        svc = self._TwitterStreamService(None, None)
        svc.set_disconnect_callback(lambda s, r: called.append((s, r)))
        self.assertEqual(called, [])
        failure = Failure(Exception())
        svc.connection_lost(failure)
        self.assertEqual(called, [(svc, failure)])

    def test_disconnect_callback_None(self):
        """
        The disconnect callback should not be called if it is unset.

        It's hard to assert that something /doesn't/ happen, so we make sure we
        see something else that happens at disconnect time and assert that no
        errors were logged.
        """
        svc = self._TwitterStreamService(None, None)
        svc._stream_response = 'foo'
        self.assertEqual(svc.disconnect_callback, None)
        failure = Failure(Exception())
        svc.connection_lost(failure)
        self.assertEqual([], self.flushLoggedErrors())

    def test_HTTP_500_initial_reconnect_delay(self):
        """
        The first HTTP error response should set the initial reconnect delay to
        one second.
        """
        d = Deferred()
        svc = self._TwitterStreamService(lambda: d, None)
        svc.clock = Clock()
        svc.startService()

        self.assertEqual(svc.reconnect_delay, 0)
        d.callback(FakeResponse(None, 500))
        self.assertEqual(svc.reconnect_delay, 1)

    def test_HTTP_500_second_reconnect_delay(self):
        """
        An HTTP error response when we already have a reconnect delay should
        double the delay and attempt to reconnect again.
        """
        d = Deferred()
        svc = self._TwitterStreamService(lambda: d, None)
        svc.clock = Clock()
        svc.reconnect_delay = 1
        svc.startService()

        self.assertEqual(svc.reconnect_delay, 1)
        d.callback(FakeResponse(None, 500))
        self.assertEqual(svc.reconnect_delay, 2)

    def test_HTTP_500_max_reconnect_delay(self):
        """
        The reconnect delay should never go over the maximum of ten minutes.
        """
        d = Deferred()
        svc = self._TwitterStreamService(lambda: d, None)
        svc.clock = Clock()
        svc.reconnect_delay = 60 * 60 * 24
        svc.startService()

        self.assertEqual(svc.reconnect_delay, 60 * 60 * 24)
        d.callback(FakeResponse(None, 500))
        self.assertEqual(svc.reconnect_delay, 60 * 10)

    def test_HTTP_500_schedules_reconnect(self):
        """
        An HTTP error should schedule a reconnection attempt.
        """
        d = Deferred()
        svc = self._TwitterStreamService(lambda: d, None)
        svc.clock = Clock()
        svc.startService()

        d.callback(FakeResponse(None, 500))
        self.assertEqual(svc._connect, svc._reconnect_delayedcall.func)

    def test_HTTP_500_calls_disconnect_callback(self):
        """
        An HTTP error should schedule a reconnection attempt.
        """
        from txtwitter.error import TwitterAPIError
        d = Deferred()
        called = []
        svc = self._TwitterStreamService(lambda: d, None)
        svc.set_disconnect_callback(lambda s, r: called.append(r))
        svc.clock = Clock()
        svc.startService()

        d.callback(FakeResponse(None, 500))
        [failure] = called
        self.assertEqual(TwitterAPIError, type(failure.value))

    def test_reconnect(self):
        """
        A reconnect should wait for the required amount of time and then
        attempt to connect again.
        """
        d1 = Deferred()
        d2 = Deferred()
        connect_deferreds = [d1, d2]
        called = []
        svc = self._TwitterStreamService(
            lambda: connect_deferreds.pop(0), None)
        svc.set_connect_callback(lambda s: called.append(s))
        svc.clock = Clock()
        svc.startService()
        d1.callback(FakeResponse(None, 500))
        self.assertEqual([], called)

        svc.clock.advance(svc.reconnect_delay)
        self.assertEqual(svc._reconnect_delayedcall, None)
        d2.callback(FakeResponse(None))
        self.assertEqual([svc], called)

    def test_rate_limit_initial_reconnect_delay(self):
        """
        The first HTTP rate limit response should set the reconnect delay to
        one minute.
        """
        d = Deferred()
        svc = self._TwitterStreamService(lambda: d, None)
        svc.clock = Clock()
        svc.startService()

        self.assertEqual(svc.reconnect_delay, 0)
        d.callback(FakeResponse(None, 420))
        self.assertEqual(svc.reconnect_delay, 60)

    def test_rate_limit_initial_reconnect_delay_existing_delay(self):
        """
        The first HTTP rate limit response should set the reconnect delay to
        one minute if there is an existing delay less than this.
        """
        d = Deferred()
        svc = self._TwitterStreamService(lambda: d, None)
        svc.clock = Clock()
        svc.reconnect_delay = 16
        svc.startService()

        d.callback(FakeResponse(None, 420))
        self.assertEqual(svc.reconnect_delay, 60)

    def test_rate_limit_second_reconnect_delay(self):
        """
        The second HTTP rate limit response should double the reconnect delay.
        """
        d = Deferred()
        svc = self._TwitterStreamService(lambda: d, None)
        svc.clock = Clock()
        svc.reconnect_delay = 60
        svc.startService()

        d.callback(FakeResponse(None, 420))
        self.assertEqual(svc.reconnect_delay, 120)

    def test_stop_service_not_started(self):
        """
        Stopping an unstarted service should do nothing.
        """
        svc = self._TwitterStreamService(lambda: None, None)
        self.assertEqual(svc.running, False)
        svc.stopService()
        self.assertEqual(svc.running, False)

    def test_stop_service_connecting(self):
        """
        Stopping a connecting service should cancel the connection.
        """
        d = Deferred()
        called = []
        svc = self._TwitterStreamService(lambda: d, None)
        svc.set_disconnect_callback(lambda s, r: called.append(r))
        svc.startService()
        self.assertEqual(svc.running, True)

        svc.stopService()
        self.assertEqual(svc.running, False)
        self.assertEqual(called, [])

    def test_stop_service_connected(self):
        """
        Stopping a connected service should close the connection cleanly.
        """
        d = Deferred()
        called = []
        svc = self._TwitterStreamService(lambda: d, None)
        svc.set_disconnect_callback(lambda s, r: called.append(r))
        svc.startService()
        d.callback(FakeResponse(None))
        self.assertEqual(svc.running, True)

        svc.stopService()
        self.assertEqual(svc.running, False)
        [failure] = called
        self.assertEqual(ResponseDone, type(failure.value))

    def test_stop_service_pending_reconnect(self):
        """
        Stopping a service with a pending reconnect should cancel the
        reconnect.
        """
        d = Deferred()
        called = []
        svc = self._TwitterStreamService(lambda: d, None)
        svc.set_disconnect_callback(lambda s, r: called.append(r))
        svc.clock = Clock()
        svc.startService()
        d.callback(FakeResponse(None, 500))
        self.assertEqual(svc.running, True)
        self.assertNotEqual(svc._reconnect_delayedcall, None)
        self.assertEqual(len(called), 1)
        self.assertNotEqual(svc.reconnect_delay, 0)

        svc.stopService()
        self.assertEqual(svc.running, False)
        self.assertEqual(svc._reconnect_delayedcall, None)
        self.assertEqual(svc.reconnect_delay, 0)



from StringIO import StringIO

from twisted.internet.defer import inlineCallbacks
from twisted.trial.unittest import TestCase
from twisted.web.client import FileBodyProducer, readBody
from twisted.web.http_headers import Headers


def from_fake_agent(name):
    @property
    def prop(self):
        from txtwitter.tests import fake_agent
        return getattr(fake_agent, name)
    return prop


class TestFakeAgent(TestCase):
    _FakeAgent = from_fake_agent('FakeAgent')
    _FakeResponse = from_fake_agent('FakeResponse')

    def test_unexpected_request(self):
        agent = self._FakeAgent()
        return self.assertFailure(agent.request('GET', 'foo'), AssertionError)

    @inlineCallbacks
    def test_no_params(self):
        agent = self._FakeAgent()
        fake_resp = object()
        agent.add_expected_request('GET', 'foo', {}, fake_resp)
        resp = yield agent.request('GET', 'foo')
        self.assertEqual(resp, fake_resp)

    @inlineCallbacks
    def test_uri_params(self):
        agent = self._FakeAgent()
        fake_resp = object()
        agent.add_expected_request('GET', 'foo', {'a': 'b'}, fake_resp)
        resp = yield agent.request('GET', 'foo?a=b')
        self.assertEqual(resp, fake_resp)

    @inlineCallbacks
    def test_body_params(self):
        agent = self._FakeAgent()
        fake_resp = object()
        agent.add_expected_request('POST', 'foo', {'a': 'b'}, fake_resp)
        resp = yield agent.request(
            'POST', 'foo', Headers({
                'Content-Type': ['application/x-www-form-urlencoded'],
            }), FileBodyProducer(StringIO('a=b')))
        self.assertEqual(resp, fake_resp)

    def test_response_static(self):
        resp = self._FakeResponse('foo', 400)
        body = self.successResultOf(readBody(resp))
        self.assertEqual(body, 'foo')
        self.assertEqual(resp.code, 400)

    def test_response_dynamic(self):
        resp = self._FakeResponse(None)
        self.assertEqual(resp.code, 200)
        d = readBody(resp)
        self.assertNoResult(d)
        resp.deliver_data('lin')
        self.assertNoResult(d)
        resp.deliver_data('e 1\nline 2\n')
        self.assertNoResult(d)
        resp.finished()
        body = self.successResultOf(d)
        self.assertEqual(body, 'line 1\nline 2\n')

    def test_response_dynamic_delayed(self):
        resp = self._FakeResponse(None)
        self.assertEqual(resp.code, 200)
        resp.deliver_data('line 1\nli')
        d = readBody(resp)
        self.assertNoResult(d)
        resp.deliver_data('ne 2\n')
        self.assertNoResult(d)
        resp.finished()
        body = self.successResultOf(d)
        self.assertEqual(body, 'line 1\nline 2\n')



from StringIO import StringIO
from urlparse import parse_qsl, urlsplit, urlunsplit

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.python.failure import Failure
from twisted.web.client import ResponseDone
from twisted.web.http import PotentialDataLoss, RESPONSES


class FakeTransport(object):
    disconnecting = False

    def __init__(self, fake_response):
        self._fake_response = fake_response

    def stopProducing(self):
        self._fake_response.finished(Failure(PotentialDataLoss()))


class FakeResponse(object):
    finished_callback = None
    _protocol = None

    def __init__(self, body, code=200):
        self.code = code
        if code == 420:
            self.phrase = 'Rate Limited'
        else:
            self.phrase = RESPONSES[code]

        if body is None:
            self._body = ''
            self._finished = False
        else:
            self._body = body
            self._finished = True

    def deliver_data(self, data):
        if self._protocol is None:
            self._body += data
        else:
            self._protocol.dataReceived(data)

    def finished(self, reason=None):
        if reason is None:
            reason = Failure(ResponseDone("Response body fully received"))
        if self.finished_callback is not None:
            self.finished_callback(reason)
        self._protocol.connectionLost(reason)

    def deliverBody(self, protocol):
        self._protocol = protocol
        protocol.makeConnection(FakeTransport(self))
        self._protocol.dataReceived(self._body)
        self._body = None
        if self._finished:
            self.finished()


class FakeAgent(object):
    def __init__(self):
        self.expected_requests = {}

    def add_expected_request(self, method, uri, params, response):
        key = (method, urlsplit(uri).geturl(), tuple(sorted(params.items())))
        self.expected_requests[key] = response

    @inlineCallbacks
    def request(self, method, uri, headers=None, bodyProducer=None):
        scheme, netloc, path, query, fragment = urlsplit(uri)
        uri = urlunsplit([scheme, netloc, path, '', ''])
        params = parse_qsl(query)

        if bodyProducer is not None:
            ctypes = headers.getRawHeaders('Content-Type')
            if ctypes == ['application/x-www-form-urlencoded']:
                consumer = StringIO()
                yield bodyProducer.startProducing(consumer)
                params.extend(parse_qsl(consumer.getvalue()))

        key = (method, uri, tuple(sorted(params)))
        assert key in self.expected_requests, (
            "Request key not found: %s" % (key,))

        returnValue(self.expected_requests[key])



import json

from twisted.application.service import Service
from twisted.internet.defer import CancelledError
from twisted.protocols.basic import LineOnlyReceiver
from twisted.protocols.policies import TimeoutMixin
from twisted.python.failure import Failure
from twisted.web.client import ResponseDone
from twisted.web.http import PotentialDataLoss

from txtwitter.error import RateLimitedError, TwitterAPIError


class TwitterStreamProtocol(LineOnlyReceiver, TimeoutMixin):
    def __init__(self, service):
        self.service = service

    def lineReceived(self, line):
        if line:
            self.service.delegate(json.loads(line))

    def connectionLost(self, reason):
        self.service.connection_lost(reason)


class TwitterStreamService(Service):
    """
    Streaming API service.

    This service handles reconnection, but does not backfill.

    From Twitter's API docs, regarding reconnections:
        <https://dev.twitter.com/docs/streaming-apis/connecting>

        Once an established connection drops, attempt to reconnect immediately.
        If the reconnect fails, slow down your reconnect attempts according to
        the type of error experienced:

         * Back off linearly for TCP/IP level network errors. These problems
           are generally temporary and tend to clear quickly. Increase the
           delay in reconnects by 250ms each attempt, up to 16 seconds.

         * Back off exponentially for HTTP errors for which reconnecting would
           be appropriate. Start with a 5 second wait, doubling each attempt,
           up to 320 seconds.

         * Back off exponentially for HTTP 420 errors. Start with a 1 minute
           wait and double each attempt. Note that every HTTP 420 received
           increases the time you must wait until rate limiting will no longer
           will be in effect for your account.

    For now, we just do an exponential backoff starting at one second and
    doubling every time we reconnect to a maximum of ten minutes. For explicit
    rate limiting, we start at 30 seconds instead of one second.
    """

    RECONNECT_DELAY_INITIAL = 1
    RECONNECT_DELAY_RATE_LIMIT = 30  # This gets doubled the first time.
    RECONNECT_DELAY_MULTIPLIER = 2
    RECONNECT_DELAY_MAX = 60 * 10

    clock = None

    _connect_d = None
    _stream_response = None
    _stream_protocol = None
    _reconnect_delayedcall = None

    connect_callback = None
    disconnect_callback = None
    reconnect_delay = 0

    def __init__(self, connect_func, delegate):
        self.connect_func = connect_func
        self.delegate = delegate

    def startService(self):
        Service.startService(self)

        if self.clock is None:
            from twisted.internet import reactor
            self.clock = reactor

        self._connect()

    def stopService(self):
        Service.stopService(self)
        if self._stream_protocol is not None:
            self._stream_protocol.transport.stopProducing()
        if self._reconnect_delayedcall is not None:
            self._reconnect_delayedcall.cancel()
            self._reconnect_delayedcall = None
        if self._connect_d is not None:
            self._connect_d.addErrback(lambda f: f.trap(CancelledError))
            self._connect_d.cancel()
            self._connect_d = None
        self.reconnect_delay = 0

    def connection_lost(self, reason):
        self._stream_response = None
        self._stream_protocol = None
        if reason.check(PotentialDataLoss):
            reason = Failure(ResponseDone())
        if self.disconnect_callback is not None:
            self.disconnect_callback(self, reason)
        self._reconnect()

    def set_connect_callback(self, callback):
        self.connect_callback = callback

    def set_disconnect_callback(self, callback):
        self.disconnect_callback = callback

    def _setup_stream(self, response):
        self._connect_d = None
        if response.code != 200:
            self._handle_HTTP_error(response)
            return

        self.reconnect_delay = self.RECONNECT_DELAY_INITIAL
        self._stream_response = response
        self._stream_protocol = TwitterStreamProtocol(self)
        response.deliverBody(self._stream_protocol)
        if self.connect_callback is not None:
            self.connect_callback(self)

    def _handle_HTTP_error(self, response):
        if response.code == 420:
            # We've been rate-limited.
            if self.reconnect_delay < self.RECONNECT_DELAY_RATE_LIMIT:
                self.reconnect_delay = self.RECONNECT_DELAY_RATE_LIMIT
            self.connection_lost(Failure(RateLimitedError(response.code)))
        else:
            # General HTTP error.
            self.connection_lost(Failure(TwitterAPIError(response.code)))

    def _connect(self):
        self._reconnect_delayedcall = None
        self._connect_d = self.connect_func()
        self._connect_d.addCallback(self._setup_stream)

    def _reconnect(self):
        if not self.running:
            return

        self._update_reconnect_delay()
        self._reconnect_delayedcall = self.clock.callLater(
            self.reconnect_delay, self._connect)

    def _update_reconnect_delay(self):
        if self.reconnect_delay < self.RECONNECT_DELAY_INITIAL:
            self.reconnect_delay = self.RECONNECT_DELAY_INITIAL
        else:
            self.reconnect_delay *= self.RECONNECT_DELAY_MULTIPLIER
        if self.reconnect_delay > self.RECONNECT_DELAY_MAX:
            self.reconnect_delay = self.RECONNECT_DELAY_MAX



import os.path

from setuptools import setup, find_packages


def readme():
    path = os.path.join(os.path.dirname(__file__), 'README.rst')
    return open(path, 'r').read()


setup(
    name="txTwitter",
    version="0.1.3a",
    url='https://github.com/jerith/txTwitter',
    license='MIT',
    description="A Twisted-based client library for Twitter's API.",
    long_description=readme(),
    author='Jeremy Thurgood',
    author_email='firxen@gmail.com',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['Twisted>=13.1.0', 'oauthlib', 'pyOpenSSL'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Twisted',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
